import warnings
from pathlib import Path
from typing import Callable, Iterable, Optional

import numpy as np
import pandas as pd
from aislib.misc_utils import get_logger
from eir.setup.input_setup_modules.setup_omics import read_bim
from joblib import Parallel, delayed
from statsmodels.base.model import LikelihoodModelResults
from statsmodels.formula import api as smf
from statsmodels.tools.sm_exceptions import ConvergenceWarning

logger = get_logger(name=__name__)

warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
warnings.filterwarnings(action="ignore", category=RuntimeWarning)


def get_allele_effects(
    df_inputs: pd.DataFrame,
    df_target: pd.DataFrame,
    df_genotype_missing: pd.DataFrame,
    bim_file: Path,
    target_type: str,
) -> pd.DataFrame:
    target_name = df_target.columns[0]
    df_combined = pd.concat(objs=[df_target, df_inputs], axis=1)

    df_bim = read_bim(bim_file_path=str(bim_file))
    rs_ids = [i for i in df_inputs.columns if not i.startswith("COVAR_")]
    allele_maps = get_snp_allele_maps(df_bim=df_bim, snp_ids=rs_ids)

    df_results = get_all_linear_results(
        df=df_combined,
        df_genotype_missing=df_genotype_missing,
        target_name=target_name,
        allele_maps=allele_maps,
        target_type=target_type,
    )

    if "P>|t|" in df_results.columns:
        df_results = df_results.rename(columns={"P>|t|": "p_value"})
    elif "P>|z|" in df_results.columns:
        df_results = df_results.rename(columns={"P>|z|": "p_value"})
    else:
        raise ValueError("Could not find p value column.")

    return df_results


def get_all_linear_results(
    df: pd.DataFrame,
    df_genotype_missing: pd.DataFrame,
    target_name: str,
    allele_maps: dict[str, dict[str, str]],
    target_type: str,
) -> pd.DataFrame:
    logger.info("Gathering all linear allele effect results for '%s'.", target_name)

    covar_columns = [i for i in df.columns if i.startswith("COVAR_")]

    parallel_worker = Parallel(n_jobs=-1)
    all_results = parallel_worker(
        delayed(_compute_single_snp_effect_wrapper)(
            df=df,
            df_genotype_missing=df_genotype_missing,
            target_name=target_name,
            allele_maps=allele_maps,
            snp=col,
            model=target_type,
            covar_columns=covar_columns,
        )
        for col in df.columns
        if col != target_name and not col.startswith("COVAR_")
    )
    all_results = [i for i in all_results if i is not None]

    if len(all_results) == 0:
        return pd.DataFrame()

    return pd.concat(all_results)


def _compute_single_snp_effect_wrapper(
    df: pd.DataFrame,
    df_genotype_missing: pd.DataFrame,
    target_name: str,
    snp: str,
    allele_maps: dict[str, dict[str, str]],
    model: str,
    covar_columns: list[str],
) -> Optional[pd.DataFrame]:
    formula = build_basic_fit_formula(
        target=target_name,
        snp=snp,
        covar_columns=covar_columns,
    )

    df_cur = df[[target_name, snp, *covar_columns]]
    df_cur_no_na = df_cur[df_genotype_missing[snp] == 0]

    n_per_group_dict = df_cur_no_na[snp].value_counts().to_dict()
    n_per_group_dict = {
        key: n_per_group_dict.get(idx, 0)
        for idx, key in enumerate(["REF", "HET", "ALT"])
    }

    fit_func = get_statsmodels_fit_function(target_type=model)

    try:
        result = fit_func(formula=formula, data=df_cur_no_na).fit(disp=0)
        if hasattr(result, "mle_retvals"):
            if not result.mle_retvals["converged"]:
                raise RuntimeError(f"Model did not converge on '{snp}'.")
        elif not np.all(np.isfinite(result.normalized_cov_params)):
            raise RuntimeError(f"Model did not converge on '{snp}'.")

        result_summary = result.summary2().tables[1]
        if not np.all(np.isfinite(result_summary)):
            raise RuntimeError(f"Model resulted in non-finite values on '{snp}'.")

    except Exception as exception:
        logger.info(
            "Failed OLS on '%s' due to exception '%s'. Skipping.",
            snp,
            exception,
        )
        return

    df_result = build_df_from_basic_results(
        results=result,
        allele_maps=allele_maps,
        snp=snp,
        n_per_group_dict=n_per_group_dict,
    )

    return df_result


def get_statsmodels_fit_function(target_type: str) -> Callable:
    if target_type == "regression":
        func = smf.ols
    elif target_type == "classification":
        func = smf.logit
    else:
        raise ValueError()

    return func


def build_basic_fit_formula(target: str, snp: str, covar_columns: list[str]) -> str:
    formula = f"Q('{target}') ~ C(Q('{snp}'))"

    for covar in covar_columns:
        formula += f" + Q('{covar}')"

    return formula


def get_snp_allele_maps(
    df_bim: pd.DataFrame,
    snp_ids: Iterable[str],
) -> dict[str, dict[str, str]]:
    snp_allele_maps = {}
    for rs_id in snp_ids:
        snp_allele_maps[rs_id] = get_snp_allele_nucleotide_map(
            df_bim=df_bim,
            rs_id=rs_id,
        )

    return snp_allele_maps


def get_snp_allele_nucleotide_map(df_bim: pd.DataFrame, rs_id: str) -> dict[str, str]:
    nucleotide_map = {}

    ref = df_bim[df_bim["VAR_ID"] == rs_id]["REF"].item()
    alt = df_bim[df_bim["VAR_ID"] == rs_id]["ALT"].item()

    nucleotide_map["REF"] = ref * 2
    nucleotide_map["HET"] = f"{ref}{alt}"
    nucleotide_map["ALT"] = alt * 2

    return nucleotide_map


def build_df_from_basic_results(
    results: LikelihoodModelResults,
    allele_maps: dict[str, dict[str, str]],
    snp: str,
    n_per_group_dict: dict[str, int],
) -> pd.DataFrame:
    coef = results.params
    std_err = results.bse
    p_values = results.pvalues
    conf_int = results.conf_int()

    df = pd.DataFrame(
        {
            "Coefficient": coef,
            "STD ERR": std_err,
            "P>|t|": p_values,
            "0.025 CI": conf_int[0],
            "0.975 CI": conf_int[1],
        }
    )

    df.index.name = "allele"

    df_renamed = _rename_linear_regression_index(
        df_results=df,
        allele_maps=allele_maps,
        snp=snp,
    )

    df_renamed["n"] = [n_per_group_dict.get(i, np.nan) for i in df_renamed["Label"]]

    df_renamed = df_renamed.rename(
        columns={
            "coef": "Coefficient",
            "std err": "STD ERR",
            "[0.025": "0.025 CI",
            "0.975]": "0.975 CI",
        }
    )

    df_renamed["KEY"] = snp

    return df_renamed


def _rename_linear_regression_index(
    df_results: pd.DataFrame,
    allele_maps: dict[str, dict[str, str]],
    snp: str,
) -> pd.DataFrame:
    df_results_copy = df_results.copy()

    snp_allele_map = allele_maps[snp]
    cur_mapping = {}
    labels = []
    for index in df_results.index:
        if index == "Intercept":
            key = "REF"
            cur_allele = snp_allele_map[key]
            cur_mapping[index] = f"{snp} {cur_allele} (Intercept)"
            labels.append(key)
        elif "[T.1]" in index:
            key = "HET"
            cur_allele = snp_allele_map[key]
            cur_mapping[index] = f"{snp} {cur_allele}"
            labels.append(key)
        elif "[T.2]" in index:
            key = "ALT"
            cur_allele = snp_allele_map[key]
            cur_mapping[index] = f"{snp} {cur_allele}"
            labels.append(key)
        elif "Q(" in index:
            start_pos = index.find("'") + 1
            end_pos = index.find("'", start_pos)
            covar_name = index[start_pos:end_pos]

            if "[T." in index and "]" in index:
                start = index.find("[T.") + 3
                end = index.find("]")
                group = index[start:end]
                group_name = f"{covar_name} (Group {group})"
            else:
                group_name = covar_name

            cur_mapping[index] = group_name
            labels.append(group_name)
        else:
            raise ValueError(f"Unexpected index format: {index}")

    df_results_copy = df_results_copy.rename(index=cur_mapping)
    df_results_copy["Label"] = labels

    return df_results_copy



import argparse
import json
import logging
import shutil
from argparse import RawTextHelpFormatter
from copy import copy
from pathlib import Path
from typing import Any, Dict, Optional, Sequence

import luigi
import pandas as pd
from aislib.misc_utils import ensure_path_exists

from eir_auto_gp.multi_task.analysis.run_analysis import RunAnalysisWrapper
from eir_auto_gp.preprocess.converge import ParseDataWrapper
from eir_auto_gp.preprocess.gwas_pre_selection import validate_geno_data_path
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)

luigi_logger = logging.getLogger("luigi")
luigi_logger.setLevel(logging.INFO)


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter)

    parser.add_argument(
        "--genotype_data_path",
        type=str,
        required=True,
        help="Root path to raw genotype data to be processed\n"
        "(e.g., containing my_data.bed, my_data.fam, my_data.bim).\n"
        "For this example, this parameter should be\n"
        "'/path/to/raw/genotype/data/'.\n"
        "Note that the file names are not included in this path,\n"
        "only the root folder. The file names are inferred, and\n"
        "*only one* set of files is expected.",
    )

    parser.add_argument(
        "--genotype_processing_chunk_size",
        type=int,
        default=1000,
        help="Chunk size for processing genotype data. Increasing"
        "this value will increase the memory usage, but will"
        "likely speed up the processing.",
    )

    parser.add_argument(
        "--label_file_path",
        type=str,
        required=True,
        help="File path to label file with tabular inputs and labels to predict.",
    )

    parser.add_argument(
        "--only_data",
        action="store_true",
        required=False,
        help="If this flag is set, only the data processing step will be run.",
    )

    parser.add_argument(
        "--global_output_folder",
        type=str,
        required=False,
        help="Common root folder to save data, feature selection and modelling results"
        " in.",
    )

    parser.add_argument(
        "--output_name",
        type=str,
        default="genotype",
        help="Name used for dataset.",
    )

    parser.add_argument(
        "--pre_split_folder",
        type=str,
        required=False,
        help="If there is a pre-split folder, this will be used to\n"
        "split the data into train/val and test sets. If not,\n"
        "the data will be split randomly.\n"
        "The folder should contain the following files:\n"
        "  - train_ids.txt: List of sample IDs to use for training.\n"
        "  - test_ids.txt: List of sample IDs to use for testing.\n"
        "  - (Optional): valid_ids.txt: List of sample IDs to use for validation.\n"
        "If this option is not specified, the data will be split randomly\n"
        "into 90/10 (train+val)/test sets.",
    )

    parser.add_argument(
        "--freeze_validation_set",
        help="If this flag is set, the validation set will be frozen\n"
        "and not changed between DL training folds.\n"
        "This only has an effect if the validation set is not specified\n"
        "in as a valid_ids.txt in file the pre_split_folder.\n"
        "If this flag is not set, the validation set will be randomly\n"
        "selected from the training set each time in each DL training run fold.\n"
        "This also has an effect when GWAS is used in feature selection.\n"
        "If the validation set is not specified manually or this flag is set,\n"
        "the GWAS will be performed on the training *and* validation set.\n"
        "This might potentially inflate the results on the validation set,\n"
        "particularly if the dataset is small. To turn off this behavior,\n"
        "you can use the --no-freeze_validation_set flag.",
        default=True,
        action="store_true",
    )

    parser.add_argument(
        "--genotype_feature_selection",
        type=str,
        default="",
        help="Feature selection method to use for genotype data.\n"
        "Options are:\n"
        "  - 'random': Randomly choose 10%% of each chromosome.\n",
    )

    parser.add_argument(
        "--folds",
        type=str,
        default="0-5",
        help="Training runs / folds to run, can be a single fold (e.g. 0),\n"
        "a range of folds (e.g. 0-5), or a comma-separated list of \n"
        "folds (e.g. 0,1,2,3,4,5).",
    )

    parser.add_argument(
        "--input_cat_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of categorical columns to use as input.",
    )

    parser.add_argument(
        "--input_con_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of continuous columns to use as input.",
    )

    parser.add_argument(
        "--output_cat_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of categorical columns to use as output.",
    )

    parser.add_argument(
        "--output_con_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of continuous columns to use as output.",
    )

    parser.add_argument(
        "--output_groups",
        type=str,
        required=False,
        help="A .yaml file containing the groups for output targets, each group "
        "will use a shared output branch in the model. The file should be in the "
        "following format:\n"
        "group_1:\n"
        "  - target_1\n"
        "  - target_2\n"
        "group_2:\n"
        "  - target_3\n"
        "  - target_4\n",
    )

    parser.add_argument(
        "--n_random_output_groups",
        type=int,
        default=2,
        help="Number of random groups to create when using 'random' for output_groups.",
    )

    parser.add_argument(
        "--model_size",
        type=str,
        default="mini",
        help="Model size to use for training.",
        choices=[
            "nano",
            "mini",
            "small",
            "medium",
            "large",
            "xlarge",
        ],
    )

    parser.add_argument(
        "--data_format",
        type=str,
        default="disk",
        help="Which format to store the data in during modelling.",
        choices=["disk", "memory", "auto"],
    )

    parser.add_argument(
        "--do_test",
        action="store_true",
        help="Whether to run test set prediction.",
    )

    return parser


def get_cl_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    cl_args = parser.parse_args()

    return cl_args


def validate_label_file(
    label_file_path: str,
    input_cat_columns: list[str],
    input_con_columns: list[str],
    output_cat_columns: list[str],
    output_con_columns: list[str],
) -> None:
    if not Path(label_file_path).exists():
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find {label_file_path}."
        )

    columns = pd.read_csv(label_file_path, nrows=1).columns
    if "ID" not in columns:
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find 'ID' column."
        )

    all_columns = set(
        input_cat_columns + input_con_columns + output_cat_columns + output_con_columns
    )
    missing_columns = all_columns - set(columns)
    if len(missing_columns) > 0:
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find columns {missing_columns}."
        )


def validate_targets(
    output_con_columns: list[str], output_cat_columns: list[str], only_data: bool
) -> None:
    if not only_data and len(output_con_columns) == 0 and len(output_cat_columns) == 0:
        raise ValueError(
            "At least one output column must be specified as continuous or categorical."
        )


def validate_plink2_exists_in_path() -> None:
    if shutil.which("plink2") is None:
        raise ValueError(
            "plink2 is not installed or not in the path. "
            "Please install plink2 and try again."
        )


def validate_pre_split_folder(pre_split_folder: Optional[str]) -> None:
    if not pre_split_folder:
        return

    ids = {}
    for file in ["train_ids.txt", "valid_ids.txt", "test_ids.txt"]:
        if not Path(pre_split_folder, file).exists():
            continue
        with open(Path(pre_split_folder, file), "r") as f:
            ids[file] = set(f.read().splitlines())

    if len(ids) == 0:
        raise ValueError(
            f"Pre-split folder {pre_split_folder} is invalid. "
            f"Expected to find at least train_ids.txt and test_ids.txt."
        )

    assert "train_ids.txt" in ids
    assert "test_ids.txt" in ids

    if "valid_ids.txt" in ids:
        assert ids["valid_ids.txt"].isdisjoint(ids["train_ids.txt"])
        assert ids["valid_ids.txt"].isdisjoint(ids["test_ids.txt"])

    assert ids["train_ids.txt"].isdisjoint(ids["test_ids.txt"])


def run(cl_args: argparse.Namespace) -> None:
    validate_geno_data_path(geno_data_path=cl_args.genotype_data_path)
    validate_label_file(
        label_file_path=cl_args.label_file_path,
        input_cat_columns=cl_args.input_cat_columns,
        input_con_columns=cl_args.input_con_columns,
        output_cat_columns=cl_args.output_cat_columns,
        output_con_columns=cl_args.output_con_columns,
    )
    validate_targets(
        output_con_columns=cl_args.output_con_columns,
        output_cat_columns=cl_args.output_cat_columns,
        only_data=cl_args.only_data,
    )
    validate_plink2_exists_in_path()
    validate_pre_split_folder(pre_split_folder=cl_args.pre_split_folder)

    cl_args = parse_output_folders(cl_args=cl_args)
    cl_args = _add_pre_split_folder_if_present(cl_args=cl_args)

    data_config = build_data_config(cl_args=cl_args)
    modelling_config = build_modelling_config(cl_args=cl_args)
    analysis_config = build_analysis_config(cl_args=cl_args)
    root_task = get_root_task(
        folds=cl_args.folds,
        data_config=data_config,
        modelling_config=modelling_config,
        analysis_config=analysis_config,
    )

    luigi.build(
        tasks=[root_task],
        workers=1,
        local_scheduler=True,
        log_level="INFO",
    )


def get_root_task(
    data_config: Dict,
    folds: int,
    modelling_config: Dict,
    analysis_config: Dict,
) -> RunAnalysisWrapper | ParseDataWrapper:
    if data_config.get("only_data"):
        return ParseDataWrapper(data_config=data_config)

    return RunAnalysisWrapper(
        folds=folds,
        data_config=data_config,
        modelling_config=modelling_config,
        analysis_config=analysis_config,
    )


def main():
    parser = get_argument_parser()
    cl_args = get_cl_args(parser=parser)
    store_experiment_config(cl_args=cl_args)

    run(cl_args=cl_args)


def store_experiment_config(
    cl_args: argparse.Namespace,
) -> None:
    config_dict = vars(cl_args)

    output_folder = Path(cl_args.global_output_folder)

    ensure_path_exists(path=output_folder, is_folder=True)
    output_path = output_folder / "config.json"

    if output_path.exists():
        logger.warning(
            f"Output config file {output_path} already exists. Overwriting it."
        )

    with open(output_path, "w") as f:
        json.dump(config_dict, f, indent=4)


def parse_output_folders(cl_args: argparse.Namespace) -> argparse.Namespace:
    cl_args_copy = copy(cl_args)
    if cl_args_copy.global_output_folder:
        gof = cl_args_copy.global_output_folder.rstrip("/")
        cl_args_copy.data_output_folder = gof + "/data"
        cl_args_copy.modelling_output_folder = gof + "/modelling"
        cl_args_copy.analysis_output_folder = gof + "/analysis"
    else:
        raise ValueError(
            "Global output folder must be specified. "
            "Please specify the --global_output_folder parameter."
        )

    return cl_args_copy


def _add_pre_split_folder_if_present(cl_args: argparse.Namespace) -> argparse.Namespace:
    cl_args_copy = copy(cl_args)
    genotype_path = Path(cl_args_copy.genotype_data_path)

    id_path = genotype_path / "ids"
    if id_path.exists():
        cl_args_copy.pre_split_folder = str(genotype_path / "ids")

        found_files = [
            i.name
            for i in id_path.iterdir()
            if i.is_file() and i.name.endswith(".txt") and "_plink" not in i.name
        ]

        logger.info(
            f"Found pre-split folder '{cl_args_copy.pre_split_folder}'. "
            f"in root genotype folder. "
            f"Using files: {found_files} for respective splits."
        )

    return cl_args_copy


def build_data_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    data_keys = [
        "genotype_data_path",
        "label_file_path",
        "data_output_folder",
        "output_name",
        "only_data",
        "pre_split_folder",
        "freeze_validation_set",
        "genotype_processing_chunk_size",
        "data_format",
    ]

    base = extract_from_namespace(namespace=cl_args, keys=data_keys)
    base["output_format"] = "deeplake"

    return base


def build_modelling_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    modelling_keys = [
        "modelling_output_folder",
        "input_cat_columns",
        "input_con_columns",
        "output_cat_columns",
        "output_con_columns",
        "output_groups",
        "n_random_output_groups",
        "genotype_feature_selection",
        "model_size",
        "do_test",
    ]

    return extract_from_namespace(namespace=cl_args, keys=modelling_keys)


def build_analysis_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    analysis_keys = [
        "analysis_output_folder",
    ]

    return extract_from_namespace(namespace=cl_args, keys=analysis_keys)


def extract_from_namespace(
    namespace: argparse.Namespace, keys: Sequence[str]
) -> Dict[str, Any]:
    return {key: getattr(namespace, key) for key in keys}


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Dict, Iterable

import numpy as np
import pandas as pd
from aislib.misc_utils import ensure_path_exists, get_logger
from matplotlib import pyplot as plt

logger = get_logger(name=__name__)


def get_snp_allele_maps(
    df_acts: pd.DataFrame, rs_ids: Iterable[str]
) -> Dict[str, Dict[str, str]]:
    snp_allele_maps = {}
    for rs_id in rs_ids:
        snp_allele_maps[rs_id] = get_snp_allele_nucleotide_map(
            df_acts=df_acts, rs_id=rs_id
        )

    return snp_allele_maps


def get_snp_allele_nucleotide_map(df_acts: pd.DataFrame, rs_id: str) -> Dict[str, str]:
    nucleotide_map = {}

    ref = df_acts[df_acts["VAR_ID"] == rs_id]["REF"].item()
    alt = df_acts[df_acts["VAR_ID"] == rs_id]["ALT"].item()

    nucleotide_map["REF"] = ref * 2
    nucleotide_map["HET"] = f"{ref}{alt}"
    nucleotide_map["ALT"] = alt * 2

    return nucleotide_map


def plot_top_snps(
    df: pd.DataFrame,
    p_value_threshold: float,
    top_n: int,
    output_dir: Path,
) -> None:
    ensure_path_exists(path=output_dir, is_folder=True)

    df_allele_effects = _parse_df_allele_effects(df_allele_effects=df)
    top_snps = _get_snps_with_strongest_effects(
        df_allele_effects=df_allele_effects,
        top_n=top_n,
    )

    for snp in top_snps:
        df_snp = df.loc[snp]

        df_total_effect = _compute_total_effect(df=df_snp)

        if np.any(df_total_effect["p_value"] > p_value_threshold):
            logger.debug(
                "SNP '%s' has p-value > %f. Skipping allele plot.",
                snp,
                p_value_threshold,
            )
            continue

        if len(df_total_effect) < 3:
            logger.debug(
                "SNP '%s' has less than 3 genotypes. Skipping allele plot.",
                snp,
            )
            continue

        plt.figure()
        plt.errorbar(
            x=df_total_effect.index.get_level_values("Genotype"),
            y=df_total_effect["Coefficient"],
            yerr=[
                abs(df_total_effect["Coefficient"] - df_total_effect["0.025 CI"]),
                abs(df_total_effect["0.975 CI"] - df_total_effect["Coefficient"]),
            ],
            fmt="o",
            capsize=5,
        )

        plt.title(f"SNP: {snp}")
        plt.xlabel("Genotype")
        plt.ylabel("Total Effect Size")
        plt.grid(True)

        plt.tight_layout()

        plt.savefig(output_dir / f"{snp}.pdf")
        plt.close()


def _parse_df_allele_effects(df_allele_effects: pd.DataFrame) -> pd.DataFrame:
    index_df = _get_index_df(df_allele_effects=df_allele_effects)

    df_allele_effects.reset_index(drop=True, inplace=True)

    cols = ["SNP", "Genotype", "Is_Intercept"]
    df_allele_effects[cols] = index_df[cols]
    df_allele_effects.set_index(keys=["SNP", "Genotype"], inplace=True)

    return df_allele_effects


def _get_index_df(df_allele_effects: pd.DataFrame) -> pd.DataFrame:
    index_df = df_allele_effects.index.to_frame(index=False)
    index_df["Is_Intercept"] = index_df["allele"].str.contains("Intercept")
    index_df["allele"] = index_df["allele"].str.replace(" (Intercept)", "")
    index_df[["SNP", "Genotype"]] = index_df["allele"].str.split(" ", expand=True)

    return index_df


def _get_snps_with_strongest_effects(
    df_allele_effects: pd.DataFrame, top_n: int
) -> pd.DataFrame:
    df_no_intercepts = df_allele_effects[~df_allele_effects["Is_Intercept"]]
    snp_strength = df_no_intercepts.groupby(level=0)["Coefficient"].apply(
        lambda x: np.abs(x).sum()
    )

    top_snps = snp_strength.nlargest(top_n).index

    return top_snps


def _compute_total_effect(
    df: pd.DataFrame,
) -> pd.DataFrame:
    row_intercept = df[df["Is_Intercept"]]
    assert len(row_intercept) == 1
    intercept_value = row_intercept["Coefficient"].item()
    intercept_index = row_intercept.index

    for index in df.index.drop(intercept_index):
        df.loc[index, "Coefficient"] += intercept_value
        df.loc[index, "0.025 CI"] += intercept_value
        df.loc[index, "0.975 CI"] += intercept_value

    return df



import subprocess
from pathlib import Path

from aislib.misc_utils import get_logger

logger = get_logger(name=__name__)


def run_prepare_data(
    final_genotype_data_path: str,
    output_folder: str,
    array_chunk_size: int,
) -> Path:
    logger.info("Running data generation pipeline for EIR.")

    command = [
        "plink_pipelines",
        "--raw_data_path",
        final_genotype_data_path,
        "--output_folder",
        output_folder,
        "--output_format",
        "disk",
        "--array_chunk_size",
        str(array_chunk_size),
    ]

    subprocess.run(command, check=True)

    output_path = get_prepared_folder_path(output_folder=output_folder)

    return output_path


def get_prepared_folder_path(output_folder: str) -> Path:
    return Path(output_folder, "processed/full_inds/full_chrs/encoded_outputs")



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
from aislib.misc_utils import get_logger
from eir.data_load.data_source_modules.deeplake_ops import load_deeplake_dataset
from eir.setup.input_setup_modules import setup_omics as eir_setup_omics
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, StandardScaler

from eir_auto_gp.single_task.modelling.dl_feature_selection import (
    get_dl_gwas_top_n_snp_list_df,
    get_dl_top_n_snp_list_df,
)
from eir_auto_gp.single_task.modelling.gwas_bo_feature_selection import (
    get_gwas_top_n_snp_list_df,
)

logger = get_logger(name=__name__)


@dataclass()
class DataPaths:
    experiment_config: Path
    train_data_path: Path
    test_data_path: Path
    train_labels_path: Path
    test_labels_path: Path
    snp_bim_path: Path
    split_ids_path: Optional[Path]
    dl_attribution_path: Path
    gwas_attribution_path: Optional[Path]
    analysis_output_path: Path


def build_data_paths(run_dir: Path) -> DataPaths:
    gwas_folder = run_dir / "feature_selection/gwas_output/"
    if gwas_folder.exists():
        gwas_attributions = next(
            i
            for i in gwas_folder.iterdir()
            if "glm.logistic" in i.name or "glm.linear" in i.name
        )
        assert gwas_attributions.exists()
    else:
        gwas_attributions = None

    config_path = run_dir / "config.json"
    config_dict = json.load(config_path.open("r"))

    data_dir = config_dict["data_output_folder"]
    if data_dir:
        data_dir = Path(data_dir)
    else:
        data_dir = run_dir / "data"

    ids_path = data_dir / "ids"
    if not ids_path.exists():
        ids_path = None

    paths = DataPaths(
        experiment_config=run_dir / "config.json",
        train_data_path=data_dir / "genotype/final/train",
        test_data_path=data_dir / "genotype/final/test",
        train_labels_path=data_dir / "tabular/final/labels_train.csv",
        test_labels_path=data_dir / "tabular/final/labels_test.csv",
        split_ids_path=ids_path,
        snp_bim_path=data_dir / "genotype/processed/parsed_files/data_final.bim",
        dl_attribution_path=run_dir
        / "feature_selection/snp_importance/dl_attributions.csv",
        gwas_attribution_path=gwas_attributions,
        analysis_output_path=run_dir / "analysis/post_analysis/",
    )

    return paths


@dataclass()
class ExperimentInfo:
    input_cat_columns: list[str]
    input_con_columns: list[str]
    all_input_columns: list[str]
    output_cat_columns: list[str]
    output_con_columns: list[str]
    target_type: str


@dataclass()
class ModelData:
    df_genotype_input: pd.DataFrame
    df_tabular_input: pd.DataFrame
    df_target: pd.DataFrame

    df_genotype_nan_mask: pd.DataFrame


def get_subset_indices_and_names(
    bim_file_path: Path,
    dl_attributions_path: Optional[Path],
    gwas_attributions_path: Optional[Path],
    top_n_snps: int,
) -> tuple[np.ndarray, list[str]]:
    df_bim = eir_setup_omics.read_bim(bim_file_path=str(bim_file_path))
    validate_file_paths(dl_path=dl_attributions_path, gwas_path=gwas_attributions_path)

    df_dl_attributions: Optional[pd.DataFrame] = None
    df_gwas_attributions: Optional[pd.DataFrame] = None

    if dl_attributions_path is not None and dl_attributions_path.exists():
        logger.info("DL attributions %s exist.", dl_attributions_path)
        df_dl_attributions = pd.read_csv(filepath_or_buffer=dl_attributions_path)

    if gwas_attributions_path is not None and gwas_attributions_path.exists():
        logger.info("GWAS attributions %s exist.", gwas_attributions_path)
        df_gwas_attributions = _read_gwas_attributions(
            gwas_attributions_path=gwas_attributions_path
        )

    match (df_dl_attributions, df_gwas_attributions):
        case (pd.DataFrame(), None):
            df_top_snps = get_dl_top_n_snp_list_df(
                df_attributions=df_dl_attributions,
                df_bim=df_bim,
                top_n_snps=top_n_snps,
            )
        case (None, pd.DataFrame()):
            df_top_snps = get_gwas_top_n_snp_list_df(
                df_gwas=df_gwas_attributions,
                top_n_snps=top_n_snps,
            )
        case (pd.DataFrame(), pd.DataFrame()):
            df_dl_gwas = df_dl_attributions.join(other=df_gwas_attributions)
            df_top_snps = get_dl_gwas_top_n_snp_list_df(
                df_dl_gwas=df_dl_gwas,
                top_n_snps=top_n_snps,
            )
        case _:
            raise ValueError("Both DL and GWAS attributions are None.")

    top_snps_list_importance_order = df_top_snps["SNP"].tolist()
    top_snps_list_ordered = _ensure_alignment_of_top_snps_with_genetic_position(
        df_bim=df_bim, top_snps_list=top_snps_list_importance_order
    )

    _check_bim(df_bim=df_bim, top_snps_list=top_snps_list_ordered)

    _ensure_all_snps_present(df_bim=df_bim, top_snps_list=top_snps_list_ordered)

    subset_indices = eir_setup_omics._setup_snp_subset_indices(
        df_bim=df_bim,
        snps_to_subset=top_snps_list_ordered,
    )

    _ensure_order_of_snps_matches(
        df_bim=df_bim,
        subset_indices=subset_indices,
        top_snps_list=top_snps_list_ordered,
    )

    return subset_indices, top_snps_list_ordered


def _ensure_all_snps_present(df_bim: pd.DataFrame, top_snps_list: list[str]) -> None:
    assert all(
        snp in df_bim["VAR_ID"].values for snp in top_snps_list
    ), "All top SNPs should be present in df_bim"


def _ensure_order_of_snps_matches(
    df_bim: pd.DataFrame, subset_indices: np.ndarray, top_snps_list: list[str]
) -> None:
    sorted_snps_in_subset = df_bim.loc[subset_indices, "VAR_ID"].tolist()
    assert (
        sorted_snps_in_subset == top_snps_list
    ), "The order of SNPs in subset_indices does not match top_snps_list"


def _ensure_alignment_of_top_snps_with_genetic_position(
    df_bim: pd.DataFrame, top_snps_list: list[str]
) -> list[str]:
    snp_to_position = dict(zip(df_bim["VAR_ID"], df_bim.index))
    top_snps_list_sorted = sorted(
        top_snps_list, key=lambda x: snp_to_position.get(x, float("inf"))
    )

    return top_snps_list_sorted


def validate_file_paths(dl_path: Optional[Path], gwas_path: Optional[Path]) -> None:
    if (dl_path is None or not dl_path.exists()) and (
        gwas_path is None or not gwas_path.exists()
    ):
        raise ValueError(
            f"Neither {dl_path} nor {gwas_path} exist. "
            "Post-analysis requires at least one of these files to exist."
        )


def _read_gwas_attributions(gwas_attributions_path: Path) -> pd.DataFrame:
    df_gwas = pd.read_csv(filepath_or_buffer=gwas_attributions_path, sep="\t")
    df_gwas = df_gwas.rename(columns={"ID": "VAR_ID"})
    df_gwas = df_gwas.set_index("VAR_ID")
    df_gwas = df_gwas.rename(columns={"P": "GWAS P-VALUE"})
    df_gwas = df_gwas[["GWAS P-VALUE"]]

    return df_gwas


def _check_bim(df_bim: pd.DataFrame, top_snps_list: list[str]) -> None:
    assert df_bim["VAR_ID"].nunique() == df_bim.shape[0]
    assert pd.Series(top_snps_list).isin(df_bim["VAR_ID"]).all()


def set_up_model_data(
    genotype_input_path: Path,
    labels_input_path: Path,
    experiment_info: ExperimentInfo,
    genotype_indices_to_load: np.ndarray,
    top_snps_list: list[str],
    data_name: str,
) -> ModelData:
    df_genotype_input = load_deeplake_samples_into_df(
        genotype_input_path=genotype_input_path,
        genotype_indices_to_load=genotype_indices_to_load,
        top_snps_list=top_snps_list,
    )

    df_tabular_input, df_target = load_tabular_data_into_df(
        labels_input_path=labels_input_path,
        experiment_info=experiment_info,
    )

    df_genotype_input.sort_index(inplace=True)
    df_tabular_input.sort_index(inplace=True)
    df_target.sort_index(inplace=True)

    df_target, df_genotype_input, df_tabular_input = _handle_missing_target_values(
        df_target=df_target,
        df_genotype_input=df_genotype_input,
        df_tabular_input=df_tabular_input,
        data_name=data_name,
    )

    df_tabular_input = _log_and_replace_inf(
        df=df_tabular_input,
        data_name=f"{data_name} tabular input",
    )
    df_genotype_input = _log_and_replace_inf(
        df=df_genotype_input,
        data_name=f"{data_name} genotype input",
    )

    assert df_genotype_input.shape[0] == df_target.shape[0] == df_tabular_input.shape[0]
    assert df_genotype_input.index.equals(df_target.index)
    assert df_genotype_input.index.equals(df_tabular_input.index)

    df_genotype_test_nan_mask = df_genotype_input.isnull().astype(int)
    model_data = ModelData(
        df_genotype_input=df_genotype_input,
        df_tabular_input=df_tabular_input,
        df_target=df_target,
        df_genotype_nan_mask=df_genotype_test_nan_mask,
    )

    return model_data


def load_deeplake_samples_into_df(
    genotype_input_path: Path,
    genotype_indices_to_load: np.ndarray,
    top_snps_list: list[str],
) -> pd.DataFrame:
    deeplake_ds = load_deeplake_dataset(data_source=str(genotype_input_path))
    ids = []
    genotype_arrays = []

    for deeplake_sample in deeplake_ds:
        sample_id = deeplake_sample["ID"].text()
        sample_genotype = deeplake_sample["genotype"].numpy()
        sample_genotype_subset = sample_genotype[:, genotype_indices_to_load]

        array_maxed = sample_genotype_subset.argmax(0).astype(np.float32)
        array_maxed[array_maxed == 3] = np.nan

        ids.append(sample_id)
        genotype_arrays.append(array_maxed)

    genotype_array = np.stack(genotype_arrays, axis=0)
    df_genotype = pd.DataFrame(data=genotype_array, columns=top_snps_list, index=ids)
    df_genotype.index.name = "ID"
    df_genotype.index = df_genotype.index.astype(str)

    return df_genotype


def load_tabular_data_into_df(
    labels_input_path: Path,
    experiment_info: ExperimentInfo,
) -> tuple[pd.DataFrame, pd.DataFrame]:

    ei = experiment_info

    all_cols = list(
        ["ID"]
        + ei.input_cat_columns
        + ei.input_con_columns
        + ei.output_cat_columns
        + ei.output_con_columns,
    )

    df_labels = pd.read_csv(
        filepath_or_buffer=labels_input_path,
        usecols=all_cols,
        index_col="ID",
    )
    df_labels.index = df_labels.index.astype(str)

    input_cols = ei.input_cat_columns + ei.input_con_columns
    df_input = df_labels[input_cols]

    target_cols = ei.output_cat_columns + ei.output_con_columns
    df_target = df_labels[target_cols]
    assert df_target.shape[1] == 1

    df_input.columns = ["COVAR_" + col for col in df_input.columns]

    return df_input, df_target


def _handle_missing_target_values(
    df_target: pd.DataFrame,
    df_genotype_input: pd.DataFrame,
    df_tabular_input: pd.DataFrame,
    data_name: str,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    logger.info("Handling missing/Inf target values, target shape: %s", df_target.shape)

    df_target = _log_and_replace_inf(df=df_target, data_name=f"{data_name} target")

    df_target = df_target.dropna()
    df_genotype_input = df_genotype_input.loc[df_target.index]
    df_tabular_input = df_tabular_input.loc[df_target.index]

    assert df_genotype_input.shape[0] == df_target.shape[0] == df_tabular_input.shape[0]

    logger.info("After dropping missing values, target shape: %s", df_target.shape)

    return df_target, df_genotype_input, df_tabular_input


def process_split_model_data(
    split_model_data: "SplitModelData",
    task_type: str,
) -> "SplitModelData":
    transformers = _set_up_transformers(
        df_train_tab_input=split_model_data.train.df_tabular_input,
        df_train_tab_target=split_model_data.train.df_target,
        task_type=task_type,
    )

    numerical_columns_ = transformers.numerical_columns
    categorical_columns_ = transformers.categorical_columns

    df_tabular_combined = pd.concat(
        [
            split_model_data.train.df_tabular_input,
            split_model_data.val.df_tabular_input,
            split_model_data.test.df_tabular_input,
        ]
    )

    continuous_imputer, categorical_imputer = setup_tabular_imputers(
        df_train=split_model_data.train.df_tabular_input,
        df_combined=df_tabular_combined,
        categorical_columns=categorical_columns_,
        numerical_columns=numerical_columns_,
    )

    categorical_maps = compute_categorical_maps(
        df_combined=df_tabular_combined,
        categorical_columns=categorical_columns_,
    )

    df_genotype_combined = pd.concat(
        [
            split_model_data.train.df_genotype_input,
            split_model_data.val.df_genotype_input,
            split_model_data.test.df_genotype_input,
        ]
    )
    genotype_imputer = setup_genotype_imputers(
        df_train=split_model_data.train.df_genotype_input,
        df_combined=df_genotype_combined,
    )

    def process_model_data(
        model_data: ModelData,
    ) -> ModelData:
        df_tabular_input_encoded = model_data.df_tabular_input
        if len(df_tabular_input_encoded.columns) > 0:

            df_tabular_input_imputed = apply_tabular_imputers(
                df=df_tabular_input_encoded,
                continuous_imputer=continuous_imputer,
                categorical_imputer=categorical_imputer,
                categorical_columns=categorical_columns_,
                numerical_columns=numerical_columns_,
            )

            df_tabular_input_encoded = apply_one_hot_encoding(
                df_base=df_tabular_input_imputed,
                category_maps=categorical_maps,
                drop_first=True,
            )

            input_scaler = transformers.input_scaler
            if input_scaler is not None:
                numerical_columns = transformers.numerical_columns
                transformed_numerical = input_scaler.transform(
                    df_tabular_input_encoded[numerical_columns]
                )
                df_tabular_input_encoded[numerical_columns] = transformed_numerical

        df_genotype_encoded = apply_genotype_imputers(
            df=model_data.df_genotype_input,
            genotype_imputer=genotype_imputer,
        )
        df_genotype_encoded = df_genotype_encoded.astype(int)

        target_data = model_data.df_target.values
        if task_type == "classification":
            label_encoder = transformers.target_label_encoder
            target_data_1d = target_data.ravel()
            target_data = label_encoder.transform(target_data_1d)

        df_target_standardized = pd.DataFrame(
            data=target_data,
            columns=model_data.df_target.columns,
            index=model_data.df_target.index,
        )

        return ModelData(
            df_genotype_input=df_genotype_encoded,
            df_tabular_input=df_tabular_input_encoded,
            df_target=df_target_standardized,
            df_genotype_nan_mask=model_data.df_genotype_nan_mask,
        )

    processed = SplitModelData(
        train=process_model_data(model_data=split_model_data.train),
        val=process_model_data(model_data=split_model_data.val),
        test=process_model_data(model_data=split_model_data.test),
        transformers=transformers,
    )

    validate_processed_split_model_data(split_model_data=processed)

    return processed


def compute_categorical_maps(
    df_combined: pd.DataFrame, categorical_columns: list[str]
) -> dict[str, list[str]]:
    category_maps = {
        col: sorted(df_combined[col].dropna().unique().tolist())
        for col in categorical_columns
    }
    return category_maps


def apply_one_hot_encoding(
    df_base: pd.DataFrame,
    category_maps: dict[str, list[str]],
    drop_first: bool = True,
) -> pd.DataFrame:

    df = df_base.copy()
    for col, categories in category_maps.items():
        for category in categories:
            df[f"{col}_{category}"] = np.array(df[col] == category).astype(int)
        if drop_first and len(categories) > 1:
            drop_category = categories[0]
            df = df.drop(f"{col}_{drop_category}", axis=1)

    return df.drop(columns=category_maps.keys())


def validate_processed_split_model_data(split_model_data: "SplitModelData") -> None:
    train = split_model_data.train
    val = split_model_data.val
    test = split_model_data.test

    train_tab = train.df_tabular_input
    val_tab = val.df_tabular_input
    test_tab = test.df_tabular_input

    assert train_tab.shape[1] == val_tab.shape[1] == test_tab.shape[1]
    assert train_tab.columns.equals(val_tab.columns)
    assert train_tab.columns.equals(test_tab.columns)

    train_gen = train.df_genotype_input
    val_gen = val.df_genotype_input
    test_gen = test.df_genotype_input

    assert train_gen.shape[1] == val_gen.shape[1] == test_gen.shape[1]
    assert train_gen.columns.equals(val_gen.columns)
    assert train_gen.columns.equals(test_gen.columns)

    train_target = train.df_target
    val_target = val.df_target
    test_target = test.df_target

    assert train_target.shape[1] == val_target.shape[1] == test_target.shape[1]
    assert train_target.columns.equals(val_target.columns)
    assert train_target.columns.equals(test_target.columns)


def setup_genotype_imputers(
    df_train: pd.DataFrame,
    df_combined: pd.DataFrame,
) -> Optional[SimpleImputer]:
    categorical_imputer = None
    if df_combined.isnull().any().any():
        categorical_imputer = SimpleImputer(strategy="most_frequent")
        missing_info = df_combined.isnull().sum()
        logger.info(
            f"Missing values in genotype columns before imputation:\n{missing_info}.\n"
            f"These missing values will be imputed for model training, but "
            f"skipped during effect analysis."
        )
        categorical_imputer.fit(df_train)

    return categorical_imputer


def apply_genotype_imputers(
    df: pd.DataFrame,
    genotype_imputer: Optional[SimpleImputer],
) -> pd.DataFrame:
    if genotype_imputer:
        imputed_values = genotype_imputer.transform(X=df)
        df = pd.DataFrame(data=imputed_values, columns=df.columns, index=df.index)

    return df


def setup_tabular_imputers(
    df_train: pd.DataFrame,
    df_combined: pd.DataFrame,
    categorical_columns: list,
    numerical_columns: list,
) -> tuple[Optional[SimpleImputer], Optional[SimpleImputer]]:
    continuous_imputer = None
    categorical_imputer = None

    logger.info(
        "Checking for missing values and setting up imputation based "
        "on training data. Imputation based on training set statistics will "
        "be applied to training, validation and test sets."
    )

    if numerical_columns:
        logger.info(f"Setting up continuous imputer for columns: {numerical_columns}")
        continuous_imputer = SimpleImputer(strategy="mean")
        if df_combined[numerical_columns].isnull().any().any():
            missing_info = df_combined[numerical_columns].isnull().sum()
            logger.info(
                f"Missing values in numerical columns "
                f"before imputation:\n{missing_info}"
            )
        continuous_imputer.fit(df_train[numerical_columns])

    if categorical_columns:
        logger.info(
            f"Setting up categorical imputer for columns: {categorical_columns}"
        )
        categorical_imputer = SimpleImputer(strategy="most_frequent")
        if df_combined[categorical_columns].isnull().any().any():
            missing_info = df_combined[categorical_columns].isnull().sum()
            logger.info(
                f"Missing values in categorical columns before"
                f" imputation:\n{missing_info}"
            )
        categorical_imputer.fit(df_train[categorical_columns])

    return continuous_imputer, categorical_imputer


def apply_tabular_imputers(
    df: pd.DataFrame,
    continuous_imputer: Optional[SimpleImputer],
    categorical_imputer: Optional[SimpleImputer],
    categorical_columns: list[str],
    numerical_columns: list[str],
) -> pd.DataFrame:
    if continuous_imputer and numerical_columns:
        df[numerical_columns] = continuous_imputer.transform(X=df[numerical_columns])

    if categorical_imputer and categorical_columns:
        df[categorical_columns] = categorical_imputer.transform(
            X=df[categorical_columns]
        )

    return df


@dataclass()
class FitTransformers:
    input_scaler: Optional[StandardScaler]
    target_label_encoder: Optional[LabelEncoder]
    numerical_columns: list[str]
    categorical_columns: list[str]


def _set_up_transformers(
    df_train_tab_input: pd.DataFrame,
    df_train_tab_target: pd.DataFrame,
    task_type: str,
) -> FitTransformers:
    numerical_columns = df_train_tab_input.select_dtypes(include=[np.number]).columns
    categorical_columns = df_train_tab_input.select_dtypes(
        include=[
            "object",
            "category",
        ]
    ).columns

    if len(numerical_columns) > 0:
        input_scaler = StandardScaler()
        input_scaler.fit(X=df_train_tab_input[numerical_columns])
    else:
        input_scaler = None

    if task_type == "classification":
        target_transformer = LabelEncoder()
        target_transformer.fit(df_train_tab_target.values.ravel())
    elif task_type == "regression":
        target_transformer = None
    else:
        raise ValueError()

    return FitTransformers(
        input_scaler=input_scaler,
        target_label_encoder=target_transformer,
        numerical_columns=numerical_columns.tolist(),
        categorical_columns=categorical_columns.tolist(),
    )


@dataclass()
class SplitModelData:
    train: ModelData
    val: ModelData
    test: ModelData

    transformers: Optional[FitTransformers] = None


def set_up_split_model_data(
    data_paths: DataPaths,
    experiment_info: ExperimentInfo,
    top_snps: int,
) -> SplitModelData:
    subset_indices, top_snps_list = get_subset_indices_and_names(
        bim_file_path=data_paths.snp_bim_path,
        dl_attributions_path=data_paths.dl_attribution_path,
        gwas_attributions_path=data_paths.gwas_attribution_path,
        top_n_snps=top_snps,
    )

    train_and_valid_data = set_up_model_data(
        genotype_input_path=data_paths.train_data_path,
        labels_input_path=data_paths.train_labels_path,
        experiment_info=experiment_info,
        genotype_indices_to_load=subset_indices,
        top_snps_list=top_snps_list,
        data_name="Train and valid",
    )

    train_ids, valid_ids, _ = maybe_gather_ids(split_ids_path=data_paths.split_ids_path)

    train_data, valid_data = split_model_data_object(
        model_data=train_and_valid_data,
        eval_size=0.1,
        ids_train=train_ids,
        ids_val=valid_ids,
    )

    test_data = set_up_model_data(
        genotype_input_path=data_paths.test_data_path,
        labels_input_path=data_paths.test_labels_path,
        experiment_info=experiment_info,
        genotype_indices_to_load=subset_indices,
        top_snps_list=top_snps_list,
        data_name="Test",
    )

    split_model_data_raw = SplitModelData(
        train=train_data,
        val=valid_data,
        test=test_data,
        transformers=None,
    )
    validate_split_model_data(split_model_data=split_model_data_raw)

    split_model_data_processed: SplitModelData = process_split_model_data(
        split_model_data=split_model_data_raw,
        task_type=experiment_info.target_type,
    )
    validate_split_model_data(split_model_data=split_model_data_processed)

    return split_model_data_processed


def maybe_gather_ids(
    split_ids_path: Optional[Path],
) -> tuple[Optional[list], Optional[list], Optional[list]]:
    if split_ids_path is None:
        return None, None, None

    train_file = split_ids_path / "train_ids.txt"
    val_file = split_ids_path / "valid_ids.txt"
    test_file = split_ids_path / "test_ids.txt"

    train_ids = list(train_file.read_text().split())

    val_ids = []
    if val_file.exists():
        val_ids = list(val_file.read_text().split())

    test_ids = list(test_file.read_text().split())

    return train_ids, val_ids, test_ids


def _log_and_replace_inf(df: pd.DataFrame, data_name: str) -> pd.DataFrame:
    numeric_cols = df.select_dtypes(include=[np.number])
    inf_mask = np.isinf(numeric_cols)

    if inf_mask.any().any():
        inf_summary = inf_mask.sum()
        logger.error(
            f"Found Inf values in '{data_name}': {inf_summary}. "
            f"Replacing Inf values with NaN."
        )
        for col in numeric_cols.columns[inf_mask.any()]:
            df[col] = df[col].replace([np.inf, -np.inf], np.nan)

    return df


def validate_split_model_data(split_model_data: "SplitModelData") -> None:
    train_ids = split_model_data.train.df_genotype_input.index
    val_ids = split_model_data.val.df_genotype_input.index
    test_ids = split_model_data.test.df_genotype_input.index

    assert len(set(train_ids) & set(val_ids)) == 0
    assert len(set(train_ids) & set(test_ids)) == 0
    assert len(set(val_ids) & set(test_ids)) == 0

    for split in [split_model_data.train, split_model_data.val, split_model_data.test]:
        assert split.df_genotype_input.index.equals(split.df_tabular_input.index)
        assert split.df_genotype_input.index.equals(split.df_target.index)


def parse_target_type(
    output_cat_columns: list[str], output_con_columns: list[str]
) -> str:
    if len(output_cat_columns) > 0:
        assert len(output_con_columns) == 0
        target_type = "classification"
    elif len(output_con_columns) > 0:
        assert len(output_cat_columns) == 0
        target_type = "regression"
    else:
        raise ValueError()

    return target_type


def split_model_data_object(
    model_data: ModelData,
    ids_train: list[str],
    ids_val: list[str],
    eval_size: float = 0.2,
) -> tuple[ModelData, ModelData]:
    df_input = pd.concat(
        objs=[model_data.df_genotype_input, model_data.df_tabular_input],
        axis=1,
    )

    if not ids_val:
        logger.info("Randomly splitting data into train and validation sets.")
        x_train, x_val, y_train, y_val = train_test_split(
            df_input,
            model_data.df_target,
            test_size=eval_size,
            random_state=42,
        )
    else:
        logger.info("Using provided IDs to split data into train and validation sets.")
        index_set = set(df_input.index)

        valid_ids_train = [id_ for id_ in ids_train if id_ in index_set]
        valid_ids_val = [id_ for id_ in ids_val if id_ in index_set]

        logger.info(f"Train IDs: {len(valid_ids_train)}/{len(ids_train)} available.")
        logger.info(f"Validation IDs: {len(valid_ids_val)}/{len(ids_val)} available.")

        x_train = df_input.loc[valid_ids_train]
        x_val = df_input.loc[valid_ids_val]
        y_train = model_data.df_target.loc[valid_ids_train]
        y_val = model_data.df_target.loc[valid_ids_val]

    df_genotype_input_train = x_train[model_data.df_genotype_input.columns]
    df_genotype_train_nan_mask = df_genotype_input_train.isnull().astype(int)
    train_data = ModelData(
        df_genotype_input=df_genotype_input_train,
        df_tabular_input=x_train[model_data.df_tabular_input.columns],
        df_target=y_train,
        df_genotype_nan_mask=df_genotype_train_nan_mask,
    )

    df_genotype_input_val = x_val[model_data.df_genotype_input.columns]
    val_data = ModelData(
        df_genotype_input=df_genotype_input_val,
        df_tabular_input=x_val[model_data.df_tabular_input.columns],
        df_target=y_val,
        df_genotype_nan_mask=df_genotype_input_val.isnull().astype(int),
    )

    return train_data, val_data


def extract_experiment_info_from_config(config_path: Path) -> ExperimentInfo:
    config_dict = json.load(config_path.open("r"))

    input_cat_columns = config_dict["input_cat_columns"]
    input_con_columns = config_dict["input_con_columns"]
    all_input_columns = input_cat_columns + input_con_columns

    output_cat_columns = config_dict["output_cat_columns"]
    output_con_columns = config_dict["output_con_columns"]

    target_type = parse_target_type(
        output_cat_columns=output_cat_columns,
        output_con_columns=output_con_columns,
    )

    experiment_info = ExperimentInfo(
        input_cat_columns=input_cat_columns,
        input_con_columns=input_con_columns,
        all_input_columns=all_input_columns,
        output_cat_columns=output_cat_columns,
        output_con_columns=output_con_columns,
        target_type=target_type,
    )

    return experiment_info



import warnings
from itertools import combinations
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from aislib.misc_utils import get_logger
from eir.setup.input_setup_modules.setup_omics import read_bim
from joblib import Parallel, delayed
from statsmodels.base.model import LikelihoodModelResults
from statsmodels.tools.sm_exceptions import ConvergenceWarning

from eir_auto_gp.post_analysis.effect_analysis.genotype_effects import (
    get_snp_allele_maps,
    get_statsmodels_fit_function,
)

logger = get_logger(name=__name__)

warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
warnings.filterwarnings(action="ignore", category=RuntimeWarning)


def get_interaction_effects(
    df_inputs: pd.DataFrame,
    df_target: pd.DataFrame,
    df_genotype_missing: pd.DataFrame,
    bim_file: Path,
    target_type: str,
    allow_within_chr_interaction: bool,
    min_interaction_pair_distance: int,
) -> pd.DataFrame:
    target_name = df_target.columns[0]
    df_combined = pd.concat(objs=[df_target, df_inputs], axis=1)

    df_bim = read_bim(bim_file_path=str(bim_file))
    snp_ids = [i for i in df_inputs.columns if not i.startswith("COVAR_")]
    allele_maps = get_snp_allele_maps(df_bim=df_bim, snp_ids=snp_ids)

    df_results = compute_interactions(
        df=df_combined,
        target_name=target_name,
        df_genotype_missing=df_genotype_missing,
        allele_maps=allele_maps,
        target_type=target_type,
        p_threshold=None,
        df_bim=df_bim,
        allow_within_chr_interaction=allow_within_chr_interaction,
        min_interaction_pair_distance=min_interaction_pair_distance,
    )

    return df_results


def compute_interactions(
    df: pd.DataFrame,
    target_name: str,
    df_genotype_missing: pd.DataFrame,
    allele_maps: dict[str, dict[str, str]],
    target_type: str,
    df_bim: pd.DataFrame,
    allow_within_chr_interaction: bool,
    min_interaction_pair_distance: int,
    p_threshold: Optional[Union[str, float]] = None,
) -> pd.DataFrame:
    logger.info(
        "Gathering all interaction allele effect results for '%s'.",
        target_name,
    )

    snps = [i for i in df.columns if i != target_name and not i.startswith("COVAR_")]
    covar_columns = [i for i in df.columns if i.startswith("COVAR_")]

    if p_threshold == "auto":
        n_snps = len(snps)
        n_tests = n_snps * (n_snps - 1) / 2
        p_threshold = 0.05 / n_tests

        logger.info(
            "p value threshold set to %f due to 'auto' option and %d SNPs.",
            p_threshold,
            len(snps),
        )

    snp_iter = filter_snp_pairs(
        df_bim=df_bim,
        snps=snps,
        allow_within_chr_interaction=allow_within_chr_interaction,
        min_interaction_pair_distance=min_interaction_pair_distance,
    )

    parallel_worker = Parallel(n_jobs=-1)
    all_results = parallel_worker(
        delayed(_compute_interaction_snp_effect_wrapper)(
            df=df,
            target_name=target_name,
            df_genotype_missing=df_genotype_missing,
            allele_maps=allele_maps,
            snp_1=snp_1,
            snp_2=snp_2,
            p_threshold=p_threshold,
            target_type=target_type,
            covar_columns=covar_columns,
        )
        for snp_1, snp_2 in snp_iter
    )
    all_results = [i for i in all_results if i is not None]

    if len(all_results) == 0:
        return pd.DataFrame()

    return pd.concat(all_results)


def filter_snp_pairs(
    df_bim: pd.DataFrame,
    snps: list[str],
    allow_within_chr_interaction: bool,
    min_interaction_pair_distance: int,
) -> list[tuple[str, str]]:
    snp_to_chr = df_bim.set_index("VAR_ID")["CHR_CODE"].to_dict()
    snp_to_pos = df_bim.set_index("VAR_ID")["BP_COORD"].to_dict()

    total_pairs = 0
    filtered_within_chr = 0
    accepted_pairs = 0

    filtered_pairs = []
    for snp_1, snp_2 in combinations(iterable=snps, r=2):
        total_pairs += 1
        chr_1, chr_2 = snp_to_chr.get(snp_1), snp_to_chr.get(snp_2)
        pos_1, pos_2 = snp_to_pos.get(snp_1), snp_to_pos.get(snp_2)

        if chr_1 == chr_2:
            over_min_distance = abs(pos_1 - pos_2) >= min_interaction_pair_distance
            if allow_within_chr_interaction and over_min_distance:
                filtered_pairs.append((snp_1, snp_2))
                accepted_pairs += 1
            else:
                filtered_within_chr += 1
        else:
            filtered_pairs.append((snp_1, snp_2))
            accepted_pairs += 1

    logger.info("Filtering SNP pairs for interaction effects.")
    logger.info(f"Total SNP pairs considered: {total_pairs}")
    logger.info(
        f"SNP pairs filtered out due to being on the same chromosome "
        f"but not meeting the distance criteria: {filtered_within_chr}"
    )
    logger.info(
        f"SNP pairs accepted (both within and between chromosomes): {accepted_pairs}"
    )

    return filtered_pairs


def _compute_interaction_snp_effect_wrapper(
    df: pd.DataFrame,
    target_name: str,
    df_genotype_missing: pd.DataFrame,
    snp_1: str,
    snp_2: str,
    p_threshold: Optional[float],
    allele_maps: dict[str, dict[str, str]],
    target_type: str,
    covar_columns: list[str],
) -> pd.DataFrame | None:
    formula = build_multiplicative_interaction_snp_formula(
        target=target_name,
        snp_1=snp_1,
        snp_2=snp_2,
        covar_columns=covar_columns,
    )

    df_cur = df[[target_name, snp_1, snp_2, *covar_columns]]

    mask_1 = df_genotype_missing[snp_1] == 0
    mask_2 = df_genotype_missing[snp_2] == 0
    df_cur_no_na = df_cur[mask_1 & mask_2]

    fit_func = get_statsmodels_fit_function(target_type=target_type)

    try:
        result = fit_func(formula=formula, data=df_cur_no_na).fit(disp=0)

        if hasattr(result, "mle_retvals"):
            if not result.mle_retvals["converged"]:
                raise RuntimeError("Model did not converge.")
        elif not np.all(np.isfinite(result.normalized_cov_params)):
            raise RuntimeError("Model did not converge.")

        result_summary = result.summary2().tables[1]
        if not np.all(np.isfinite(result_summary)):
            raise RuntimeError("Model resulted in non-finite values")

    except Exception as exception:
        logger.info(
            "Failed OLS on '%s' and '%s' due to exception '%s'. Skipping.",
            snp_1,
            snp_2,
            exception,
        )
        return None

    if p_threshold is not None:
        p_value = _extract_p_value_from_interaction_results(results=result)
        if p_value > p_threshold:
            logger.info(
                "Rejected interaction effect between %s and %s due to p-value %f.",
                snp_1,
                snp_2,
                p_value,
            )
            return None

    try:
        df_result = build_df_from_interaction_results(
            results=result,
            allele_maps=allele_maps,
            snp_1=snp_1,
            snp_2=snp_2,
        )
        return df_result
    except Exception as e:
        logger.info(
            "Failed parsing interaction results on '%s' and '%s' due to exception "
            "'%s'. Skipping.",
            snp_1,
            snp_2,
            e,
        )
        return None


def _extract_p_value_from_interaction_results(
    results: LikelihoodModelResults,
) -> float:
    p_values = results.pvalues
    last_row_name = p_values.index[-1]
    assert ":" in last_row_name
    p_value = p_values.iloc[-1]
    return p_value


def build_multiplicative_interaction_snp_formula(
    target: str,
    snp_1: str,
    snp_2: str,
    covar_columns: list[str],
) -> str:
    formula = (
        f"Q('{target}') ~ C(Q('{snp_1}')) + C(Q('{snp_2}')) + Q('{snp_1}'):Q('{snp_2}')"
    )

    for covar in covar_columns:
        formula += f" + Q('{covar}')"

    return formula


def build_df_from_interaction_results(
    results: LikelihoodModelResults,
    allele_maps: dict[str, dict[str, str]],
    snp_1: str,
    snp_2: str,
) -> pd.DataFrame:
    coef = results.params
    std_err = results.bse
    p_values = results.pvalues
    conf_int = results.conf_int()

    df = pd.DataFrame(
        {
            "Coefficient": coef,
            "STD ERR": std_err,
            "P>|t|": p_values,
            "0.025 CI": conf_int[0],
            "0.975 CI": conf_int[1],
        }
    )

    df.index.name = "allele"

    df.index = df.index.str.replace(r"C\(Q\(", "", regex=True)
    df.index = df.index.str.replace(r"Q\(", "", regex=True)
    df.index = df.index.str.replace(r"\)", "", regex=True)
    df.index = df.index.str.replace(r"'", "", regex=True)

    df = _rename_interaction_regression_index(
        df_results=df,
        allele_maps=allele_maps,
        snp=snp_1,
    )

    df = _rename_interaction_regression_index(
        df_results=df,
        allele_maps=allele_maps,
        snp=snp_2,
    )

    df = df.rename(index={f"{snp_1}:{snp_2}": f"{snp_1}--:--{snp_2}"})

    df["KEY"] = f"{snp_1}--:--{snp_2}"

    df = df.rename(
        columns={
            "coef": "Coefficient",
            "std err": "STD ERR",
            "[0.025": "0.025 CI",
            "0.975]": "0.975 CI",
        }
    )

    return df


def _rename_interaction_regression_index(
    df_results: pd.DataFrame,
    allele_maps: dict[str, dict[str, str]],
    snp: str,
) -> pd.DataFrame:
    snp_allele_map = allele_maps[snp]
    cur_mapping = {}
    for index in df_results.index:
        if "[T.1]" in index and snp in index:
            key = "HET"
            cur_allele = snp_allele_map[key]
            cur_mapping[index] = f"{snp} {cur_allele}"
        elif "[T.2]" in index and snp in index:
            key = "ALT"
            cur_allele = snp_allele_map[key]
            cur_mapping[index] = f"{snp} {cur_allele}"
        elif "Q(" in index:
            start_pos = index.find("'") + 1
            end_pos = index.find("'", start_pos)
            covar_name = index[start_pos:end_pos]

            if "[T." in index and "]" in index:
                start = index.find("[T.") + 3
                end = index.find("]")
                group = index[start:end]
                group_name = f"{covar_name} (Group {group})"
            else:
                group_name = covar_name

            cur_mapping[index] = group_name
        else:
            continue

    df_results = df_results.rename(index=cur_mapping)

    return df_results



from dataclasses import dataclass
from itertools import combinations
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Generator, Optional
from warnings import simplefilter

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from aislib.misc_utils import ensure_path_exists, get_logger
from sklearn.preprocessing import StandardScaler

from eir_auto_gp.post_analysis.run_complexity_analysis import (
    ModelReadyObject,
    convert_split_data_to_model_ready_object,
    train_and_evaluate_linear,
)

if TYPE_CHECKING:
    from eir_auto_gp.post_analysis.run_post_analysis import PostAnalysisObject

simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

logger = get_logger(name=__name__)


@dataclass
class StepInformationObjects:
    complexity_results: pd.DataFrame
    allele_effects: pd.DataFrame
    interaction_effects: pd.DataFrame


def _get_step_information_objects(
    post_analysis_object: "PostAnalysisObject",
    eval_set: str,
) -> StepInformationObjects:
    assert eval_set in ["valid", "test"]
    output_root = post_analysis_object.data_paths.analysis_output_path

    df_complexity = pd.read_csv(output_root / f"complexity/{eval_set}/all_results.csv")

    df_allele_effects = pd.read_csv(
        output_root / "effect_analysis/allele_effects/allele_effects.csv"
    )
    df_allele_effects = filter_snp_rows(df=df_allele_effects)

    df_interaction_effects = pd.read_csv(
        output_root / "effect_analysis/interaction_effects/interaction_effects.csv"
    )

    if len(df_interaction_effects) != 0:
        df_interaction_effects = filter_snp_rows(df=df_interaction_effects)

    return StepInformationObjects(
        complexity_results=df_complexity,
        allele_effects=df_allele_effects,
        interaction_effects=df_interaction_effects,
    )


def filter_snp_rows(df: pd.DataFrame) -> pd.DataFrame:
    df_filtered = df[~df["allele"].str.contains("COVAR")].copy()
    return df_filtered


def run_iterative_complexity_analysis(
    post_analysis_object: "PostAnalysisObject",
    n_iterative_complexity_candidates: int,
    eval_set: str,
) -> None:
    if eval_set not in ["test", "valid"]:
        raise ValueError(
            f"Unsupported eval_set: {eval_set}. Must be 'test' or 'valid'."
        )

    pao = post_analysis_object
    ei = pao.experiment_info
    sidp = _get_step_information_objects(
        post_analysis_object=pao,
        eval_set=eval_set,
    )

    outputs = ei.output_cat_columns + ei.output_con_columns
    if len(outputs) > 1:
        raise ValueError(
            "Multiple outputs detected, iterative complexity analysis "
            "only supports single output."
        )

    output_name = outputs[0]

    output_root = (
        pao.data_paths.analysis_output_path / "iterative_complexity" / eval_set
    )
    ensure_path_exists(path=output_root, is_folder=True)

    results = []

    n = n_iterative_complexity_candidates
    data_iter = get_step_training_iterator(
        post_analysis_object=pao,
        step_information_objects=sidp,
        max_one_hot=n,
        max_interaction_exe=n,
        max_interaction_gxe=n,
        max_interaction_gxg=n,
        eval_set=eval_set,
    )

    for idx, (mro, group, model_type) in enumerate(data_iter):
        train_eval_results = train_and_evaluate_linear(
            modelling_data=mro,
            target_type=ei.target_type,
            eval_set=eval_set,
            model_type="elasticnet",
            cv_use_val_split=True,
            with_fallback=True,
        )
        df_performance = train_eval_results.performance
        df_performance["model_type"] = model_type
        df_performance["group"] = group
        df_performance["step"] = idx
        results.append(df_performance)

    df_results = pd.concat(results)
    df_results = df_results.reset_index(drop=True)

    df_results.to_csv(output_root / "iterative_complexity_results.csv")

    figures_output_folder = output_root / "figures"
    ensure_path_exists(path=figures_output_folder, is_folder=True)

    plot_iterative_complexity_results(
        df_results=df_results,
        output_name=output_name,
        df_complexity_results=sidp.complexity_results,
        output_folder=figures_output_folder,
    )


def plot_iterative_complexity_results(
    df_results: pd.DataFrame,
    df_complexity_results: pd.DataFrame,
    output_name: str,
    output_folder: Path,
) -> None:
    ensure_path_exists(path=output_folder, is_folder=True)

    df_results, metric_columns = parse_metrics_changes(df=df_results)

    for metric in metric_columns:
        top_performance, line_label = get_top_performance_info(
            df_complexity_results=df_complexity_results,
            metric=metric,
        )

        fig = plot_metric_figure(
            df=df_results,
            output_name=output_name,
            metric=metric,
            top_performance=top_performance,
            line_label=line_label,
        )
        fig.savefig(output_folder / f"{metric}_change.pdf")
        plt.close(fig)

        direction = "increase" if metric != "rmse" else "decrease"
        fig_changes = plot_top_changes(
            df=df_results,
            metric=metric,
            metric_change=f"{metric}_change",
            output_name=output_name,
            direction=direction,
        )
        fig_changes.savefig(output_folder / f"{metric}_top_changes.pdf")
        plt.close(fig_changes)


def parse_metrics_changes(df: pd.DataFrame) -> tuple[pd.DataFrame, list[str]]:
    df_copy = df.copy()
    skip_cols = ["Unnamed: 0", "model_type", "step", "group"]
    metric_columns = [col for col in df_copy.columns if col not in skip_cols]
    for metric in metric_columns:
        df_copy[f"{metric}_change"] = df_copy[metric].diff().fillna(0)
    return df_copy, metric_columns


def get_top_performance_info(df_complexity_results: pd.DataFrame, metric: str):
    if metric == "rmse":
        top_row = df_complexity_results.loc[df_complexity_results[metric].idxmin()]
    else:
        top_row = df_complexity_results.loc[df_complexity_results[metric].idxmax()]

    label_components = [
        "XGB" if top_row["model_type"] == "xgboost" else "LIN",
        "GT" if top_row["include_genotype"] else "",
        "TAB" if top_row["include_tabular"] else "",
        "OH" if top_row["one_hot_encode"] else "",
    ]
    label = "-".join(filter(None, label_components))

    return top_row[metric], label


def plot_top_changes(
    df: pd.DataFrame,
    metric_change: str,
    metric: str,
    output_name: str,
    top_n: int = 3,
    direction: str = "increase",
) -> plt.Figure:
    if direction not in ["increase", "decrease"]:
        raise ValueError("direction must be 'increase' or 'decrease'")

    excluded_terms = {"Tabular", "Genotype", "Genotype + Tabular"}
    df_filtered = df[~df["model_type"].isin(excluded_terms)]

    condition = (
        f"{metric_change} > 0" if direction == "increase" else f"{metric_change} < 0"
    )

    df_relevant_changes = df_filtered.query(condition)

    sort_ascending = direction == "decrease"

    df_top_changes = (
        df_relevant_changes.sort_values(by=[metric_change], ascending=sort_ascending)
        .groupby("group")
        .head(top_n)
    )

    fig = plt.figure(figsize=(10, 6))
    sns.barplot(
        data=df_top_changes,
        x=metric_change,
        y="model_type",
        hue="group",
        orient="h",
        palette="deep",
    )
    title_suffix = "Decreases" if direction == "decrease" else "Improvements"
    xlabel_suffix = "Reduction" if direction == "decrease" else "Improvement"
    plt.title(
        f"Top {top_n} {title_suffix} in {output_name}",
        fontsize=16,
        weight="bold",
    )
    plt.xlabel(f"{metric} {xlabel_suffix}", fontsize=14)
    plt.ylabel("Model Type", fontsize=14)
    plt.tight_layout()
    plt.legend(title="Group", bbox_to_anchor=(1.05, 1), loc="upper left")

    return fig


def plot_metric_figure(
    df: pd.DataFrame,
    metric: str,
    top_performance: float,
    line_label: str,
    output_name: str,
) -> plt.Figure:
    sns.set(style="whitegrid", palette="deep")

    change_col = f"{metric}_change"
    direction = "min" if metric == "rmse" else "max"

    df_parsed = prepare_dataframe_with_other_terms(
        df=df,
        metric_change=change_col,
        metric=metric,
        top_n=3,
        direction=direction,
    )

    y_size = 0.75 * len(df_parsed["model_type"].unique())
    longest_string = max(df_parsed["model_type"], key=len)
    x_size = max(8, int(0.35 * len(longest_string)))
    plt.figure(figsize=(x_size, y_size))

    ax = sns.barplot(
        data=df_parsed,
        x=metric,
        y="model_type",
        orient="h",
        hue="group",
    )

    ax.set_title(
        f"{output_name}",
        fontsize=16,
        weight="bold",
    )
    ax.set_xlabel(f"{metric.upper()} Change", fontsize=14)
    ax.set_ylabel("Model Type", fontsize=14)
    ax.xaxis.label.set_color("black")
    ax.yaxis.label.set_color("black")
    ax.tick_params(axis="x", colors="black")
    ax.tick_params(axis="y", colors="black")

    for p, change in zip(ax.patches, df_parsed[change_col]):
        change_sign = "+" if change >= 0 else "-"
        ax.annotate(
            f"{change_sign}{abs(change):.3f}",
            xy=(p.get_width(), p.get_y() + p.get_height() / 2),
            xytext=(0, 0),
            textcoords="offset points",
            ha="left",
            va="center",
            fontsize=12,
            color="black",
            weight="bold",
        )

    line_color = "black"
    plt.axvline(
        x=top_performance,
        color=line_color,
        linestyle="--",
        linewidth=1,
    )
    plt.text(
        1.01,
        0.5,
        f"--- {line_label}",
        transform=ax.transAxes,
        color=line_color,
        va="center",
        ha="left",
        fontsize=12,
        weight="bold",
        bbox=dict(
            facecolor="white",
            alpha=1.0,
            edgecolor="black",
            boxstyle="round,pad=0.5",
        ),
    )

    ax.get_legend().remove()

    y_ticks = ax.get_yticks()
    y_labels = [label.get_text() for label in ax.get_yticklabels()]
    updated_labels = [
        "Other Terms" if "Other terms" in label else label for label in y_labels
    ]
    ax.set_yticks(y_ticks)
    ax.set_yticklabels(updated_labels)

    sns.despine(left=False, bottom=False, right=True, top=True)
    plt.tight_layout()

    return plt.gcf()


def prepare_dataframe_with_other_terms(
    df: pd.DataFrame,
    metric: str,
    metric_change: str = "r2_change",
    top_n: int = 3,
    direction: str = "max",
) -> pd.DataFrame:
    df_copy = df.copy()
    df_copy[metric] = df_copy[metric].fillna(0)

    df_processed = pd.DataFrame()
    df_copy = df_copy.sort_values(by="step")

    for group in df_copy["group"].unique():
        df_group = df_copy[df_copy["group"] == group]
        if direction == "max":
            top_contributors = df_group.nlargest(n=top_n, columns=metric_change)
        else:
            top_contributors = df_group.nsmallest(n=top_n, columns=metric_change)

        df_other_terms = pd.DataFrame()

        last_step_included = -1
        for i, row in df_group.iterrows():
            if row["step"] <= last_step_included:
                continue

            if row["model_type"] in top_contributors["model_type"].values:
                df_processed = pd.concat(
                    [df_processed, df_other_terms, pd.DataFrame([row])],
                    ignore_index=True,
                )
                df_other_terms = pd.DataFrame()

                last_step_included = row["step"]
            else:
                if df_other_terms.empty:
                    df_other_terms = pd.DataFrame([row])
                else:
                    n_rows = len(df_processed)
                    df_other_terms[metric_change] += row[metric_change]
                    df_other_terms[metric] = row[metric]
                    df_other_terms["model_type"] = f"Other terms {n_rows}"
                    df_other_terms["step"] = row["step"]

        if not df_other_terms.empty:
            df_processed = pd.concat(
                [df_processed, df_other_terms],
                ignore_index=True,
            )

    df_processed["group"] = df_processed["group"].replace(
        {
            "Tabular": "Base",
            "Genotype": "Base",
            "Genotype + Tabular": "Base",
        }
    )

    return df_processed


def get_step_training_iterator(
    post_analysis_object: "PostAnalysisObject",
    step_information_objects: StepInformationObjects,
    max_one_hot: int,
    max_interaction_exe: int,
    max_interaction_gxe: int,
    max_interaction_gxg: int,
    eval_set: str,
    include_genotype: bool = True,
    include_tabular: bool = True,
) -> Generator[tuple[ModelReadyObject, str, str], None, None]:
    """
    1. Only tabular
    2. Only genotype
    3. Genotype + tabular
    4. Genotype + tabular + tabular complex terms
    5. Iteratively one-hot encode SNPs up to max_one_hot
    6. One-hot encode all SNPs
    7. Iteratively add Tabular x Tabular interaction terms up to max_interaction_exe
    8. Iteratively add SNP x Tabular interaction terms up to max_interaction_gxe
    9. Iteratively add SNP x SNP interaction terms up to max_interaction_gxg
    """

    ei = post_analysis_object.experiment_info
    any_tabular = len(ei.all_input_columns) > 0
    run_tabular = include_tabular and any_tabular
    run_genotype = include_genotype

    if not run_tabular and not run_genotype:
        logger.warning(
            "No data to run iterative complexity analysis on as both "
            "tabular and genotype are excluded, skipping."
        )
        return

    # 1
    tabular_feature_importance = None
    running_mro = None

    if run_tabular:
        logger.info("Running iterative complexity analysis: tabular only")
        mro_tabular = convert_split_data_to_model_ready_object(
            split_model_data=post_analysis_object.modelling_data,
            include_genotype=False,
            include_tabular=True,
            one_hot_encode=False,
        )

        yield mro_tabular, "Tabular", "Tabular"

        tabular_results = train_and_evaluate_linear(
            modelling_data=mro_tabular,
            target_type=ei.target_type,
            eval_set=eval_set,
            cv_use_val_split=True,
            with_fallback=True,
        )
        tabular_feature_importance = tabular_results.feature_importance
        running_mro = mro_tabular

    # 2
    if run_genotype:
        logger.info("Running iterative complexity analysis: genotype only")
        mro_genotype = convert_split_data_to_model_ready_object(
            split_model_data=post_analysis_object.modelling_data,
            include_genotype=True,
            include_tabular=False,
            one_hot_encode=False,
        )

        yield mro_genotype, "Genotype", "Genotype"
        running_mro = mro_genotype

    # 3
    if run_tabular and run_genotype:
        logger.info("Running iterative complexity analysis: genotype + tabular")
        mro_genotype_tabular = convert_split_data_to_model_ready_object(
            split_model_data=post_analysis_object.modelling_data,
            include_genotype=True,
            include_tabular=True,
            one_hot_encode=False,
        )

        yield mro_genotype_tabular, "Genotype + Tabular", "Genotype + Tabular"
        running_mro = mro_genotype_tabular

    # 4
    if run_tabular:
        logger.info("Running iterative complexity analysis: tabular complex terms")
        mro_tabular_complex_terms, _ = get_t_term_iterator(
            post_analysis_object=post_analysis_object,
            running_mro=running_mro,
        )

        for mro_t_term, model_type in mro_tabular_complex_terms:
            yield mro_t_term, "Tabular Complex Terms", model_type
            running_mro = mro_t_term

    # 5
    if run_genotype:
        logger.info("Running iterative complexity analysis: one-hot encoding")
        one_hot_iterator = _get_one_hot_iterator(
            post_analysis_object=post_analysis_object,
            running_mro=running_mro,
            df_allele_effects=step_information_objects.allele_effects,
            max_one_hot=max_one_hot,
        )

        for cur_mro_oh, model_type in one_hot_iterator:
            yield cur_mro_oh, "OH", model_type
            running_mro = cur_mro_oh

    # 6
    if run_genotype:
        logger.info("Running iterative complexity analysis: all one-hot encoding")
        mro_all_oh, model_type = one_hot_encode_all_snps(
            running_mro=running_mro,
            df_allele_effects=step_information_objects.allele_effects,
        )

        yield mro_all_oh, "All OH", model_type
        running_mro = mro_all_oh

    # 7
    if run_tabular:
        logger.info("Running iterative complexity analysis: tabular x tabular")
        txt_iterator, _, _ = get_txt_iterator(
            running_mro=running_mro,
            post_analysis_object=post_analysis_object,
            top_n=max_interaction_exe,
        )

        if txt_iterator is not None:
            for mro_txt, model_type in txt_iterator:
                yield mro_txt, "TxT", model_type
                running_mro = mro_txt

    # 8
    if run_tabular and run_genotype:
        logger.info("Running iterative complexity analysis: genotype x tabular")
        assert tabular_feature_importance is not None
        gxt_iterator, _, _ = get_gxt_iterator(
            running_mro=running_mro,
            post_analysis_object=post_analysis_object,
            top_n=max_interaction_gxe,
            step_information_objects=step_information_objects,
            tabular_feature_importance=tabular_feature_importance,
        )

        for mro_gxt, model_type in gxt_iterator:
            yield mro_gxt, "GxT", model_type
            running_mro = mro_gxt

    # 9
    if run_genotype:
        logger.info("Running iterative complexity analysis: genotype x genotype")
        gxg_iterator = get_gxg_iterator(
            running_mro=running_mro,
            post_analysis_object=post_analysis_object,
            interaction_effects_df=step_information_objects.interaction_effects,
            top_n=max_interaction_gxg,
        )

        for cur_mro_oh, model_type in gxg_iterator:
            yield cur_mro_oh, "GxG", model_type


def _get_one_hot_iterator(
    post_analysis_object: "PostAnalysisObject",
    running_mro: ModelReadyObject,
    df_allele_effects: pd.DataFrame,
    max_one_hot: int,
) -> Generator:
    pao = post_analysis_object

    genotype_columns = pao.modelling_data.train.df_genotype_input.columns.tolist()
    df_top_non_additive_snps = _find_non_additive_snps(
        df_allele_effects=df_allele_effects,
        genotype_columns=genotype_columns,
        n=max_one_hot,
    )

    for snp in df_top_non_additive_snps["KEY"]:

        def one_hot_encode_snp(
            inputs: pd.DataFrame, targets: pd.DataFrame
        ) -> tuple[pd.DataFrame, pd.DataFrame]:
            if snp in inputs.columns:
                inputs = pd.get_dummies(
                    inputs,
                    columns=[snp],
                    prefix=snp + "_OH",
                    drop_first=True,
                )
            return inputs, targets

        running_mro = _merge_operate_and_split(
            model_ready_object=running_mro, function=one_hot_encode_snp
        )

        cur_name = f"+ OH {snp}"

        yield running_mro, cur_name


def _find_non_additive_snps(
    df_allele_effects: pd.DataFrame,
    genotype_columns: list[str],
    n: int,
) -> pd.DataFrame:
    non_additive_effects = []

    df_all_filt = df_allele_effects[
        df_allele_effects["KEY"].isin(genotype_columns)
    ].copy()

    genotype_order = {"REF": 0, "HET": 1, "ALT": 2}
    df_all_filt["genotype_order"] = df_all_filt["Label"].map(genotype_order)

    for key, group in df_all_filt.sort_values(["KEY", "genotype_order"]).groupby("KEY"):
        if len(group) == 3:
            sorted_group = group.sort_values("genotype_order")
            ref_effect, het_effect, alt_effect = sorted_group["Coefficient"].values
            non_additivity_metric = abs(alt_effect - 2 * het_effect)
            non_additive_effects.append((key, non_additivity_metric))

    non_additive_df = (
        pd.DataFrame(non_additive_effects, columns=["KEY", "Non_Additive_Metric"])
        .sort_values(by="Non_Additive_Metric", ascending=False)
        .head(n)
    )

    return non_additive_df


def one_hot_encode_all_snps(
    running_mro: ModelReadyObject, df_allele_effects: pd.DataFrame
) -> tuple:
    snps_to_encode = df_allele_effects["KEY"].unique()

    snps_to_encode = [
        snp for snp in snps_to_encode if snp in running_mro.input_train.columns
    ]

    def one_hot_encode_multiple_snps(
        inputs: pd.DataFrame, targets: pd.DataFrame
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        if snps_to_encode:
            inputs = pd.get_dummies(
                data=inputs,
                columns=snps_to_encode,
                prefix={snp: snp + "_OH" for snp in snps_to_encode},
                drop_first=True,
            )
        return inputs, targets

    updated_mro = _merge_operate_and_split(
        model_ready_object=running_mro, function=one_hot_encode_multiple_snps
    )

    operation_name = "+ All OH"

    return updated_mro, operation_name


def get_t_term_iterator(
    post_analysis_object: "PostAnalysisObject",
    running_mro: ModelReadyObject,
) -> tuple[Generator, ModelReadyObject]:
    """
    Note: For now we are only considering numerical columns here. This can be extended
          to also include the one-hot-encoded columns from categorical variables,
          by looping over:
          [i for i in running_mro.df_train.columns if i.startswith("_COVAR")]
    """

    scaler = running_mro.transformers.input_scaler
    numerical_columns = running_mro.transformers.numerical_columns

    train_ids = running_mro.input_train.index

    def generator():
        nonlocal running_mro
        for col in numerical_columns:

            def add_squared_term(
                inputs: pd.DataFrame, targets: pd.DataFrame
            ) -> tuple[pd.DataFrame, pd.DataFrame]:
                col_name = f"{col}"
                term = f"{col}^2"

                inputs[term] = inputs[col_name] ** 2

                return inputs, targets

            def add_cubed_term(
                inputs: pd.DataFrame, targets: pd.DataFrame
            ) -> tuple[pd.DataFrame, pd.DataFrame]:
                col_name = f"{col}"
                term = f"{col}^3"
                inputs[term] = inputs[col_name] ** 3

                return inputs, targets

            def add_log_term(
                inputs: pd.DataFrame,
                targets: pd.DataFrame,
            ) -> tuple[pd.DataFrame, pd.DataFrame]:
                """
                We have this kind of roundabout way of doing this because we need to
                inverse scale the column (otherwise won't work due to <=0 values
                from scaling) and then take the log, then scale again for models like
                ElasticNetCV.
                """
                col_name: str = f"{col}"
                term: str = f"log_{col_name}"

                if scaler is not None:
                    col_data = inverse_scale_column(
                        scaler=scaler,
                        inputs=inputs,
                        col_name=col_name,
                        numerical_columns=numerical_columns,
                    )

                    if np.all(col_data > 0):
                        inputs[term] = np.log(col_data)
                    else:
                        raise ValueError(
                            f"Column {col_name} has non-positive values, "
                            f"cannot take the log of this column"
                        )

                    inputs = fit_transform_on_train_only(
                        inputs=inputs,
                        train_ids=train_ids,
                        column=term,
                    )

                return inputs, targets

            def add_inverse_term(
                inputs: pd.DataFrame, targets: pd.DataFrame
            ) -> tuple[pd.DataFrame, pd.DataFrame]:
                col_name = f"{col}"
                term = f"1/{col}"

                if (inputs[col_name] == 0).any():
                    raise ValueError(
                        f"Column {col_name} contains zero values, "
                        f"cannot take the inverse of this column"
                    )
                else:
                    inputs[term] = 1 / inputs[col_name]

                return inputs, targets

            running_mro = _merge_operate_and_split(
                model_ready_object=running_mro,
                function=add_squared_term,
            )
            yield running_mro, f"+ T {col}^2"

            running_mro = _merge_operate_and_split(
                model_ready_object=running_mro,
                function=add_cubed_term,
            )
            yield running_mro, f"+ T {col}^3"

            maybe_running_mro = _merge_operate_and_split(
                model_ready_object=running_mro,
                function=add_log_term,
            )

            if isinstance(maybe_running_mro, ModelReadyObject):
                running_mro = maybe_running_mro
                yield running_mro, f"+ T log_{col}"

            maybe_running_mro = _merge_operate_and_split(
                model_ready_object=running_mro,
                function=add_inverse_term,
            )

            if isinstance(maybe_running_mro, ModelReadyObject):
                running_mro = maybe_running_mro
                yield running_mro, f"+ T 1/{col}"

    return generator(), running_mro


def inverse_scale_column(
    scaler: StandardScaler,
    inputs: pd.DataFrame,
    col_name: str,
    numerical_columns: list[str],
) -> np.ndarray:
    col_position: int = numerical_columns.index(col_name)

    mean: float = scaler.mean_[col_position]
    scale: float = scaler.scale_[col_position]

    col_data: np.ndarray = inputs[col_name].to_numpy()
    inversely_scaled_data: np.ndarray = col_data * scale + mean
    return inversely_scaled_data


def fit_transform_on_train_only(
    inputs: pd.DataFrame,
    train_ids: pd.Index,
    column: str,
) -> pd.DataFrame:
    scaler = StandardScaler()
    scaler.fit(X=inputs.loc[train_ids, [column]])
    inputs[column] = scaler.transform(X=inputs[[column]])
    return inputs


def get_txt_iterator(
    post_analysis_object: "PostAnalysisObject",
    running_mro: ModelReadyObject,
    top_n: int,
) -> tuple[Optional[Generator], pd.DataFrame, ModelReadyObject]:
    pao = post_analysis_object

    target_type = pao.experiment_info.target_type
    numerical_columns = running_mro.transformers.numerical_columns

    mro_txt_search = convert_split_data_to_model_ready_object(
        split_model_data=pao.modelling_data,
        include_genotype=False,
        include_tabular=True,
        one_hot_encode=False,
    )

    txt_candidates = _find_txt_candidates(
        model_ready_object=mro_txt_search,
        tabular_columns=numerical_columns,
        target_type=target_type,
        top_n=top_n,
    )

    if txt_candidates.empty:
        logger.warning("No TxT candidates found, skipping.")
        return None, pd.DataFrame(), running_mro

    logger.debug(
        "Top %d TxT candidates: %s", top_n, txt_candidates["interaction"].tolist()
    )

    def generator():
        nonlocal running_mro
        for idx, row in txt_candidates.iterrows():
            col1, col2 = row["interaction"].split("_x_")

            def add_interaction(
                inputs: pd.DataFrame, targets: pd.DataFrame
            ) -> tuple[pd.DataFrame, pd.DataFrame]:
                col1_name = f"{col1}"
                col2_name = f"{col2}"
                interaction_term = f"{col1_name}_x_{col2_name}"
                inputs[interaction_term] = inputs[col1_name] * inputs[col2_name]
                return inputs, targets

            running_mro = _merge_operate_and_split(
                model_ready_object=running_mro,
                function=add_interaction,
            )

            yield running_mro, f"+ TxT {col1}_x_{col2}"

    return generator(), txt_candidates, running_mro


def _find_txt_candidates(
    model_ready_object: ModelReadyObject,
    tabular_columns: list[str],
    target_type: str,
    top_n: int,
) -> pd.DataFrame:
    """
    TODO: Possibly make this just return the top_n feature important ones from
          the linear model only trained on tabular.
    """
    results = []
    mro = model_ready_object

    logger.debug(
        "Searching for top TxT candidates among %d tabular columns, "
        "testing %d combinations",
        len(tabular_columns),
        len(tabular_columns) * (len(tabular_columns) - 1) / 2,
    )

    for col1, col2 in combinations(iterable=tabular_columns, r=2):

        def add_interaction(
            inputs: pd.DataFrame,
            targets: pd.DataFrame,
        ) -> tuple[pd.DataFrame, pd.DataFrame]:
            col1_name = f"{col1}"
            col2_name = f"{col2}"
            interaction_term = f"{col1_name}_x_{col2_name}"
            inputs[interaction_term] = inputs[col1_name] * inputs[col2_name]
            return inputs, targets

        modified_mro = _merge_operate_and_split(
            model_ready_object=mro,
            function=add_interaction,
        )

        train_eval_results = train_and_evaluate_linear(
            modelling_data=modified_mro,
            target_type=target_type,
            eval_set="valid",
            cv_use_val_split=True,
            with_fallback=True,
        )
        df_performance = train_eval_results.performance
        df_performance["interaction"] = f"{col1}_x_{col2}"

        results.append(df_performance)

    if not results:
        return pd.DataFrame()

    df_results = pd.concat(results)
    metric = "r2" if target_type == "regression" else "mcc"
    df_sorted = df_results.sort_values(by=metric, ascending=False)
    return df_sorted.head(top_n)


def get_gxt_iterator(
    running_mro: ModelReadyObject,
    post_analysis_object: "PostAnalysisObject",
    step_information_objects: StepInformationObjects,
    tabular_feature_importance: pd.DataFrame,
    top_n: int,
) -> tuple[Generator, pd.DataFrame, ModelReadyObject]:
    pao = post_analysis_object
    target_type = pao.experiment_info.target_type

    mro_search = convert_split_data_to_model_ready_object(
        split_model_data=pao.modelling_data,
        include_genotype=True,
        include_tabular=True,
        one_hot_encode=False,
    )
    genotype_columns = pao.modelling_data.train.df_genotype_input.columns.tolist()
    gxt_candidates = _find_gxt_candidates(
        model_ready_object=mro_search,
        target_type=target_type,
        genotype_columns=genotype_columns,
        top_n=top_n,
        df_allele_effects=step_information_objects.allele_effects,
        df_tabular_feature_importance=tabular_feature_importance,
    )

    logger.debug(
        "Top %d GxT candidates: %s", top_n, gxt_candidates["interaction"].tolist()
    )

    def generator():
        nonlocal running_mro

        for idx, row in gxt_candidates.iterrows():
            snp_column, tab_column = row["interaction"].split("_x_")

            def add_interaction(
                inputs: pd.DataFrame, targets: pd.DataFrame
            ) -> tuple[pd.DataFrame, pd.DataFrame]:
                snp_column_additive = pd.Series(0, index=inputs.index)
                for genotype in [0, 1, 2]:
                    snp_col = f"{snp_column}_OH_{genotype}"
                    if snp_col in inputs.columns:
                        snp_column_additive += inputs[snp_col] * genotype

                interaction_term = f"{snp_column}_x_{tab_column}"

                inputs[interaction_term] = snp_column_additive * inputs[tab_column]

                return inputs, targets

            running_mro = _merge_operate_and_split(
                model_ready_object=running_mro,
                function=add_interaction,
            )

            yield running_mro, f"+ GxT {snp_column}_x_{tab_column}"

    return generator(), gxt_candidates, running_mro


def _find_gxt_candidates(
    model_ready_object: ModelReadyObject,
    target_type: str,
    top_n: int,
    genotype_columns: list[str],
    df_allele_effects: pd.DataFrame,
    df_tabular_feature_importance: pd.DataFrame,
) -> pd.DataFrame:
    """
    Here we assume we are operating on additive encodings.

    TODO: Possibly we can skip the filtering there below, this is likely just an
          issue when manual doing partial runs with top_n_snps being different
          from the original run and the current iterative complexity run.
    """
    results = []
    mro = model_ready_object

    df_allele_effects_filtered = df_allele_effects[
        df_allele_effects["KEY"].isin(genotype_columns)
    ]

    top_snp_candidates = _find_top_snp_candidates(
        df_allele_effects=df_allele_effects_filtered,
        top_n=top_n,
    )

    df_tabular_feature_importance = df_tabular_feature_importance[
        df_tabular_feature_importance["Feature"] != "Intercept"
    ]
    top_tabular_columns = df_tabular_feature_importance.head(top_n)["Feature"].tolist()

    logger.debug(
        "Estimated n(TxG) combinations: %d for top %d SNPs: %s x top %d tabular: %s",
        len(top_snp_candidates) * len(top_tabular_columns),
        len(top_snp_candidates),
        top_snp_candidates,
        len(top_tabular_columns),
        top_tabular_columns,
    )

    for snp_column in top_snp_candidates:
        for tab_column in top_tabular_columns:

            def add_interaction(
                inputs: pd.DataFrame, targets: pd.DataFrame
            ) -> tuple[pd.DataFrame, pd.DataFrame]:
                interaction_values = inputs[snp_column] * inputs[tab_column]

                interaction_name = f"{snp_column}_x_{tab_column}"
                inputs[interaction_name] = interaction_values
                return inputs, targets

            modified_mro = _merge_operate_and_split(
                model_ready_object=mro,
                function=add_interaction,
            )

            train_eval_results = train_and_evaluate_linear(
                modelling_data=modified_mro,
                target_type=target_type,
                eval_set="valid",
                cv_use_val_split=True,
                with_fallback=True,
            )
            df_performance = train_eval_results.performance
            df_performance["interaction"] = f"{snp_column}_x_{tab_column}"

            results.append(df_performance)

    df_results = pd.concat(results)
    metric = "r2" if target_type == "regression" else "mcc"
    return df_results.sort_values(by=metric, ascending=False).head(top_n)


def _find_top_snp_candidates(df_allele_effects: pd.DataFrame, top_n: int) -> list[str]:
    snp_effects_total = {}

    for snp, group in df_allele_effects.groupby("KEY"):
        snp_effects_total[snp] = group["Coefficient"].abs().sum()

    snp_effects_total_df = pd.DataFrame(
        snp_effects_total.items(), columns=["KEY", "Total_Effect"]
    )

    top_snps_list = (
        snp_effects_total_df.sort_values(by="Total_Effect", ascending=False)
        .head(top_n)["KEY"]
        .tolist()
    )

    return top_snps_list


def get_gxg_iterator(
    running_mro: ModelReadyObject,
    post_analysis_object: "PostAnalysisObject",
    interaction_effects_df: pd.DataFrame,
    top_n: int,
) -> Generator:
    """
    TODO: Possibly we can skip the filtering there below, this is likely just an
          issue when manual doing partial runs with top_n_snps being different
          from the original run and the current iterative complexity run.
    """
    mro = running_mro
    pao = post_analysis_object

    genotype_columns = pao.modelling_data.train.df_genotype_input.columns.tolist()

    if len(interaction_effects_df) == 0:
        logger.warning(
            "No interaction effects found between SNPs in the genotype"
            " columns, skipping."
        )
        return

    interaction_effects_df["SNP1"] = (
        interaction_effects_df["KEY"].str.split("--:--").str[0]
    )
    interaction_effects_df["SNP2"] = (
        interaction_effects_df["KEY"].str.split("--:--").str[1]
    )

    interaction_effects_df_filtered = interaction_effects_df[
        (interaction_effects_df["SNP1"].isin(genotype_columns))
        & (interaction_effects_df["SNP2"].isin(genotype_columns))
    ].copy()

    if len(interaction_effects_df_filtered) == 0:
        logger.warning(
            "No interaction effects found between SNPs in the genotype"
            " columns, skipping."
        )
        return

    top_gxg_candidates = _find_top_gxg_candidates(
        interaction_effects_df=interaction_effects_df_filtered,
        top_n=top_n,
    )

    for idx, row in top_gxg_candidates.iterrows():
        snp_pair = row["KEY"]
        snp1, snp2 = snp_pair.split("--:--")

        def add_interaction(
            inputs: pd.DataFrame,
            targets: pd.DataFrame,
        ) -> tuple[pd.DataFrame, pd.DataFrame]:
            snp1_additive = pd.Series(0, index=inputs.index)
            snp2_additive = pd.Series(0, index=inputs.index)

            for genotype in [0, 1, 2]:
                snp1_col = f"{snp1}_OH_{genotype}"
                if snp1_col in inputs.columns:
                    snp1_additive += inputs[snp1_col] * genotype

                snp2_col = f"{snp2}_OH_{genotype}"
                if snp2_col in inputs.columns:
                    snp2_additive += inputs[snp2_col] * genotype

            interaction_term = snp1_additive * snp2_additive
            interaction_name = f"{snp1}_x_{snp2}"
            inputs[interaction_name] = interaction_term

            return inputs, targets

        modified_mro = _merge_operate_and_split(
            model_ready_object=mro,
            function=add_interaction,
        )

        yield modified_mro, f"+ GxG {snp_pair}"


def _find_top_gxg_candidates(
    interaction_effects_df: pd.DataFrame, top_n: int
) -> pd.DataFrame:
    interaction_rows = interaction_effects_df[
        interaction_effects_df["allele"] == interaction_effects_df["KEY"]
    ].copy()

    interaction_rows["Coefficient"] = interaction_rows["Coefficient"].abs()

    top_interactions = interaction_rows.sort_values(
        by="Coefficient", ascending=False
    ).head(top_n)

    return top_interactions[["KEY", "Coefficient"]]


class FailedMergeOperateAndSplit:
    pass


def _merge_operate_and_split(
    model_ready_object: ModelReadyObject,
    function: Callable[
        [pd.DataFrame, pd.DataFrame],
        tuple[pd.DataFrame, pd.DataFrame],
    ],
) -> ModelReadyObject | FailedMergeOperateAndSplit:
    df_all_inputs = pd.concat(
        (
            model_ready_object.input_train,
            model_ready_object.input_val,
            model_ready_object.input_test,
        )
    )

    df_all_targets = pd.concat(
        (
            model_ready_object.target_train,
            model_ready_object.target_val,
            model_ready_object.target_test,
        )
    )

    try:
        df_all_inputs, df_all_targets = function(df_all_inputs, df_all_targets)
    except Exception as e:
        logger.error("Error when applying function to model ready object: %s", e)
        return FailedMergeOperateAndSplit()

    input_train = df_all_inputs.loc[model_ready_object.input_train.index]
    target_train = df_all_targets.loc[model_ready_object.target_train.index]
    assert input_train.index.equals(model_ready_object.input_train.index)
    assert target_train.index.equals(model_ready_object.target_train.index)

    input_val = df_all_inputs.loc[model_ready_object.input_val.index]
    target_val = df_all_targets.loc[model_ready_object.target_val.index]
    assert input_val.index.equals(model_ready_object.input_val.index)
    assert target_val.index.equals(model_ready_object.target_val.index)

    input_test = df_all_inputs.loc[model_ready_object.input_test.index]
    target_test = df_all_targets.loc[model_ready_object.target_test.index]
    assert input_test.index.equals(model_ready_object.input_test.index)
    assert target_test.index.equals(model_ready_object.target_test.index)

    return ModelReadyObject(
        input_train=input_train,
        target_train=target_train,
        input_val=input_val,
        target_val=target_val,
        input_test=input_test,
        target_test=target_test,
        transformers=model_ready_object.transformers,
    )



import os
import subprocess
from dataclasses import dataclass, fields
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, Iterable, Literal, Optional

import luigi
import pandas as pd
import yaml
from aislib.misc_utils import ensure_path_exists
from eir.setup.config_setup_modules.config_setup_utils import recursive_dict_inject
from eir.setup.input_setup_modules.setup_omics import read_bim

from eir_auto_gp.multi_task.modelling.configs import (
    AggregateConfig,
    get_aggregate_config,
)
from eir_auto_gp.preprocess.converge import (
    ParseDataWrapper,
    get_batch_size,
    get_dynamic_valid_size,
)
from eir_auto_gp.single_task.modelling.run_modelling import (
    get_bim_path,
    get_dataloader_workers,
    get_device,
    get_memory_dataset,
    get_samples_per_epoch,
    lines_in_file,
)
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


class RunModellingWrapper(luigi.Task):
    folds = luigi.Parameter()
    data_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()

    def requires(self):
        task_object = (
            TestSingleRun if self.modelling_config["do_test"] else TrainSingleRun
        )
        for fold in _get_fold_iterator(folds=str(self.folds)):
            yield task_object(
                fold=fold,
                data_config=self.data_config,
                modelling_config=self.modelling_config,
            )

        cleanup_tmp_files(modelling_config=self.modelling_config)

    def output(self):
        return self.input()


def cleanup_tmp_files(modelling_config: Dict[str, Any]) -> None:
    tmp_dir = Path(modelling_config["modelling_output_folder"]) / "tmp"
    if tmp_dir.exists():
        for file in tmp_dir.iterdir():
            file.unlink()
        tmp_dir.rmdir()


def _get_fold_iterator(folds: str) -> Iterable:
    if "-" in folds:
        start, end = folds.split("-")
        return range(int(start), int(end) + 1)
    elif "," in folds:
        return [int(i) for i in folds.split(",")]
    else:
        return range(int(folds))


class TestSingleRun(luigi.Task):
    fold = luigi.IntParameter()

    data_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()

    def requires(self):
        base = {
            "data": ParseDataWrapper(data_config=self.data_config),
            "train_run": TrainSingleRun(
                fold=self.fold,
                data_config=self.data_config,
                modelling_config=self.modelling_config,
            ),
        }

        return base

    def run(self):
        output_root = Path(self.output().path).parent
        ensure_path_exists(output_root, is_folder=True)

        all_target_columns = (
            self.modelling_config["output_cat_columns"]
            + self.modelling_config["output_con_columns"]
        )

        base_aggregate_config = get_aggregate_config(
            output_head="linear",
            output_groups=self.modelling_config["output_groups"],
            model_size=self.modelling_config["model_size"],
            target_columns=all_target_columns,
            output_cat_columns=self.modelling_config["output_cat_columns"],
            output_con_columns=self.modelling_config["output_con_columns"],
            n_random_groups=self.modelling_config["n_random_output_groups"],
        )

        injection_params = build_injection_params(
            fold=int(self.fold),
            data_input_dict=self.input()["data"][0],
            task="test",
            data_config=self.data_config,
            modelling_config=self.modelling_config,
            output_configs=base_aggregate_config.output_config,
        )

        injections = _get_all_dynamic_injections(
            injection_params=injection_params,
            genotype_data_path=self.data_config["genotype_data_path"],
        )

        with TemporaryDirectory() as temp_dir:
            temp_config_folder = Path(temp_dir)
            build_configs(
                injections=injections,
                aggregate_config_base=base_aggregate_config,
                output_folder=temp_config_folder,
            )

            train_run_folder = Path(self.input()["train_run"].path).parent
            base_predict_command = get_testing_string_from_config_folder(
                config_folder=temp_config_folder,
                train_run_folder=train_run_folder,
                with_labels=True,
            )
            logger.info("Testing command: %s", base_predict_command)

            base_command_split = base_predict_command.split()
            process_info = subprocess.run(
                base_command_split,
                env=dict(os.environ, EIR_SEED=str(self.fold)),
            )

            if process_info.returncode != 0:
                raise RuntimeError(
                    f"Testing failed with return code {process_info.returncode}"
                )
            testing_complete_file = Path(self.output().path)
            testing_complete_file.touch()

    def output(self):
        run_folder = Path(self.input()["train_run"].path).parent
        output_path = run_folder / "test_set_predictions/test_complete.txt"
        return luigi.LocalTarget(str(output_path))


def get_testing_string_from_config_folder(
    config_folder: Path,
    train_run_folder: Path,
    with_labels: bool,
) -> str:
    base_string = "eirpredict"
    globals_string = " --global_configs "
    inputs_string = " --input_configs "
    fusion_string = " --fusion_configs "
    output_string = " --output_configs "

    for file in config_folder.iterdir():
        if file.suffix == ".yaml":
            if "global" in file.stem:
                globals_string += " " + f"{str(file)}" + " "
            elif "input" in file.stem:
                inputs_string += " " + f"{str(file)}" + " "
            elif "fusion" in file.stem:
                fusion_string += " " + f"{str(file)}" + " "
            elif "output" in file.stem:
                output_string += " " + f"{str(file)}" + " "

    final_string = (
        base_string + globals_string + inputs_string + fusion_string + output_string
    )

    saved_models = list((train_run_folder / "saved_models").iterdir())
    assert len(saved_models) == 1, "Expected only one saved model."

    final_string += f" --model_path {saved_models[0]}"
    if with_labels:
        final_string += " --evaluate"

    test_output_folder = train_run_folder / "test_set_predictions"
    ensure_path_exists(path=test_output_folder, is_folder=True)

    final_string += f" --output_folder {test_output_folder}"

    return final_string


class TrainSingleRun(luigi.Task):
    fold = luigi.IntParameter()
    data_config = luigi.DictParameter()
    modelling_config = luigi.DictParameter()

    def requires(self):
        base = {"data": ParseDataWrapper(data_config=self.data_config)}
        for i in range(self.fold):
            base[f"train_{i}"] = TrainSingleRun(
                fold=i,
                data_config=self.data_config,
                modelling_config=self.modelling_config,
            )
        return base

    def run(self):
        output_root = Path(self.output().path).parent
        ensure_path_exists(output_root, is_folder=True)

        all_target_columns = (
            self.modelling_config["output_cat_columns"]
            + self.modelling_config["output_con_columns"]
        )

        base_aggregate_config = get_aggregate_config(
            output_head="linear",
            output_groups=self.modelling_config["output_groups"],
            model_size=self.modelling_config["model_size"],
            target_columns=all_target_columns,
            output_cat_columns=self.modelling_config["output_cat_columns"],
            output_con_columns=self.modelling_config["output_con_columns"],
            n_random_groups=self.modelling_config["n_random_output_groups"],
        )

        injection_params = build_injection_params(
            fold=int(self.fold),
            data_input_dict=self.input()["data"][0],
            task="train",
            data_config=self.data_config,
            modelling_config=self.modelling_config,
            output_configs=base_aggregate_config.output_config,
        )

        injections = _get_all_dynamic_injections(
            injection_params=injection_params,
            genotype_data_path=self.data_config["genotype_data_path"],
        )

        with TemporaryDirectory() as temp_dir:
            temp_config_folder = Path(temp_dir)
            build_configs(
                injections=injections,
                aggregate_config_base=base_aggregate_config,
                output_folder=temp_config_folder,
            )

            base_train_command = get_training_string_from_config_folder(
                config_folder=temp_config_folder
            )
            logger.info("Training command: %s", base_train_command)

            base_command_split = base_train_command.split()
            process_info = subprocess.run(
                base_command_split,
                env=dict(os.environ, EIR_SEED=str(self.fold)),
            )

            if process_info.returncode != 0:
                raise RuntimeError(
                    f"Training failed with return code {process_info.returncode}"
                )
            training_complete_file = Path(self.output().path)
            training_complete_file.touch()

    def output(self):
        modelling_output_folder = Path(self.modelling_config["modelling_output_folder"])
        fold_output_folder = modelling_output_folder / f"fold_{self.fold}"
        file_name = fold_output_folder / "completed_train.txt"
        local_target = luigi.LocalTarget(path=str(file_name))
        return local_target


@dataclass
class ModelInjectionParams:
    fold: int
    modelling_base_output_folder: str
    output_folder: str
    manual_valid_ids_file: Optional[str]
    genotype_input_source: str
    genotype_feature_selection: str
    label_file_path: str
    input_cat_columns: list[str]
    input_con_columns: list[str]
    output_cat_columns: list[str]
    output_con_columns: list[str]
    weighted_sampling_columns: list[str]
    data_format: str
    output_configs: list[dict[str, Any]]


def build_injection_params(
    fold: int,
    data_input_dict: Dict[str, luigi.LocalTarget],
    task: Literal["train", "test"],
    data_config: Dict[str, Any],
    modelling_config: Dict[str, Any],
    output_configs: list[Dict[str, Any]],
) -> ModelInjectionParams:
    weighted_sampling_columns = None

    base_output_folder = modelling_config["modelling_output_folder"]
    cur_run_output_folder = f"{base_output_folder}/fold_{fold}"

    label_file_path = data_input_dict[f"{task}_tabular"].path

    manual_valid_ids_file = get_manual_valid_ids_file(
        task=task,
        data_config=data_config,
        label_file_path=label_file_path,
        base_output_folder=base_output_folder,
    )

    params = ModelInjectionParams(
        fold=fold,
        output_folder=cur_run_output_folder,
        modelling_base_output_folder=base_output_folder,
        manual_valid_ids_file=str(manual_valid_ids_file),
        genotype_input_source=data_input_dict[f"{task}_genotype"].path,
        label_file_path=label_file_path,
        genotype_feature_selection=modelling_config["genotype_feature_selection"],
        input_cat_columns=modelling_config["input_cat_columns"],
        input_con_columns=modelling_config["input_con_columns"],
        output_cat_columns=modelling_config["output_cat_columns"],
        output_con_columns=modelling_config["output_con_columns"],
        weighted_sampling_columns=weighted_sampling_columns,
        data_format=data_config["data_format"],
        output_configs=output_configs,
    )

    return params


def get_weighted_sampling_columns(
    modelling_config: Dict[str, Any]
) -> Optional[list[str]]:
    return ["all"] if modelling_config["output_cat_columns"] else None


def get_manual_valid_ids_file(
    task: str,
    data_config: Dict[str, Any],
    label_file_path: str,
    base_output_folder: str,
) -> Optional[str]:
    if task != "train":
        return None

    valid_ids_file = Path(data_config["data_output_folder"], "ids/valid_ids.txt")
    if not valid_ids_file.exists():
        return None

    manual_valid_ids = pd.read_csv(valid_ids_file, header=None)[0].tolist()
    df_ids = pd.read_csv(label_file_path, usecols=["ID"]).dropna()
    intersected_ids = list(
        set(manual_valid_ids).intersection(set(df_ids["ID"].tolist()))
    )

    manual_valid_ids_file = Path(base_output_folder, "tmp", "valid_ids.txt")
    manual_valid_ids_file.parent.mkdir(parents=True, exist_ok=True)

    with open(manual_valid_ids_file, "w") as f:
        for id_ in intersected_ids:
            f.write(f"{id_}\n")

    return str(manual_valid_ids_file)


def build_tmp_label_file(
    label_file_path: str,
    tmp_dir: Path,
    output_cat_columns: list[str],
    output_con_columns: list[str],
    prefix_name: str,
) -> str:
    tmp_label_file_path = Path(tmp_dir) / f"{prefix_name}_label_file.csv"

    if tmp_label_file_path.exists():
        return str(tmp_label_file_path)

    df = pd.read_csv(label_file_path)

    if prefix_name == "test":
        train_file_path = Path(tmp_dir) / "train_label_file.csv"
        assert Path(train_file_path).exists(), train_file_path
        df_train = pd.read_csv(train_file_path, usecols=["ID"])
        ids_train = set(df_train["ID"].tolist())
        ids_test = set(df["ID"].tolist())
        assert len(ids_train.intersection(ids_test)) == 0

    ensure_path_exists(path=tmp_label_file_path.parent, is_folder=True)
    df.to_csv(tmp_label_file_path, index=False)

    return str(tmp_label_file_path)


def build_configs(
    aggregate_config_base: AggregateConfig,
    injections: Dict[str, Any],
    output_folder: Path,
) -> None:
    for config_field in fields(aggregate_config_base):
        config_name = config_field.name
        config = getattr(aggregate_config_base, config_name)
        if config_name in injections:

            if config_name == "output_config":
                assert isinstance(config, list)

                for output_config in config:
                    output_config_name = output_config["output_info"]["output_name"]
                    cur_to_inject = injections[config_name][output_config_name]
                    output_config = recursive_dict_inject(
                        dict_=output_config,
                        dict_to_inject=cur_to_inject,
                    )

                    validate_complete_config(config_element=output_config)

                    out_path = output_folder / f"output_{output_config_name}.yaml"

                    with open(out_path, "w") as f:
                        yaml.dump(output_config, f)

            else:
                config = recursive_dict_inject(
                    dict_=config,
                    dict_to_inject=injections[config_name],
                )
                validate_complete_config(config_element=config)
                with open(output_folder / f"{config_name}.yaml", "w") as f:
                    yaml.dump(config, f)
        else:
            continue


def validate_complete_config(config_element: dict | list | str) -> None:
    match config_element:
        case dict(config_element):
            for key, value in config_element.items():
                validate_complete_config(config_element=value)
        case list(config_element):
            for value in config_element:
                validate_complete_config(config_element=value)
        case str(config_element):
            assert config_element != "FILL"


def get_training_string_from_config_folder(config_folder: Path) -> str:
    base_string = "eirtrain"
    globals_string = " --global_configs "
    inputs_string = " --input_configs "
    fusion_string = " --fusion_configs "
    output_string = " --output_configs "

    for file in config_folder.iterdir():
        if file.suffix == ".yaml" and "_test" not in file.stem:
            if "global" in file.stem:
                globals_string += " " + f"{str(file)}" + " "
            elif "input" in file.stem:
                inputs_string += " " + f"{str(file)}" + " "
            elif "fusion" in file.stem:
                fusion_string += " " + f"{str(file)}" + " "
            elif "output" in file.stem:
                output_string += " " + f"{str(file)}" + " "

    final_string = (
        base_string + globals_string + inputs_string + fusion_string + output_string
    )

    return final_string


def _get_global_injections(
    fold: int,
    output_folder: str,
    valid_size: int,
    batch_size: int,
    manual_valid_ids_file: Optional[str],
    n_snps: int,
    n_samples: int,
    iter_per_epoch: int,
    weighted_sampling_columns: list[str],
    data_format: str,
) -> Dict[str, Any]:
    mixing_candidates = [0.0]
    cur_mixing = mixing_candidates[fold % len(mixing_candidates)]

    device = get_device()

    if data_format == "auto":
        memory_dataset = get_memory_dataset(n_snps=n_snps, n_samples=n_samples)
    elif data_format == "disk":
        memory_dataset = False
    elif data_format == "memory":
        memory_dataset = True
    else:
        raise ValueError(f"Unknown data format: '{data_format}'.")

    n_workers = get_dataloader_workers(memory_dataset=memory_dataset, device=device)
    early_stopping_buffer = min(5000, iter_per_epoch * 5)
    early_stopping_buffer = max(early_stopping_buffer, 1000)
    sample_interval = min(1000, iter_per_epoch)
    lr = _get_learning_rate(n_snps=n_snps)

    injections = {
        "basic_experiment": {
            "output_folder": output_folder,
            "device": device,
            "batch_size": batch_size,
            "valid_size": valid_size,
            "manual_valid_ids_file": manual_valid_ids_file,
            "dataloader_workers": n_workers,
            "memory_dataset": memory_dataset,
        },
        "optimization": {
            "lr": lr,
        },
        "evaluation_checkpoint": {
            "sample_interval": sample_interval,
            "checkpoint_interval": sample_interval,
        },
        "training_control": {
            "mixing_alpha": cur_mixing,
            "early_stopping_buffer": early_stopping_buffer,
            "weighted_sampling_columns": weighted_sampling_columns,
        },
    }

    return injections


def _get_learning_rate(n_snps: int) -> float:
    if n_snps < 1_000:
        lr = 1e-03
    elif n_snps < 10_000:
        lr = 5e-04
    elif n_snps < 100_000:
        lr = 2e-04
    elif n_snps < 500_000:
        lr = 1e-04
    elif n_snps < 2_000_000:
        lr = 5e-05
    else:
        lr = 1e-05

    logger.info("Setting learning rate to %f due to %d SNPs.", lr, n_snps)

    return lr


def _get_genotype_injections(
    input_source: str,
    n_snps: int,
    subset_snp_path: Optional[Path],
) -> Dict[str, Any]:
    base_snp_path = (
        Path(input_source).parent.parent / "processed/parsed_files/data_final.bim"
    )
    assert base_snp_path.exists(), f"SNP file not found at {base_snp_path}"

    kernel_width, first_kernel_expansion = get_gln_kernel_parameters(n_snps=n_snps)

    injections = {
        "input_info": {
            "input_source": input_source,
        },
        "input_type_info": {
            "snp_file": str(base_snp_path),
        },
        "model_config": {
            "model_init_config": {
                "kernel_width": kernel_width,
                "first_kernel_expansion": first_kernel_expansion,
            }
        },
    }

    if subset_snp_path:
        injections["input_type_info"]["subset_snps_file"] = str(subset_snp_path)

    return injections


def get_gln_kernel_parameters(n_snps: int) -> tuple[int, int]:
    if n_snps < 1000:
        params = 16, -4
    elif n_snps < 10000:
        params = 16, -2
    elif n_snps < 100000:
        params = 16, 1
    elif n_snps < 500000:
        params = 16, 2
    elif n_snps < 2000000:
        params = 16, 4
    else:
        params = 16, 8

    logger.info(
        "Setting kernel width to %d and first kernel expansion to %d due to %d SNPs.",
        params[0],
        params[1],
        n_snps,
    )

    return params


def _get_tabular_injections(
    input_source: str, input_cat_columns: list[str], input_con_columns: list[str]
) -> Dict[str, Any]:
    injections = {
        "input_info": {
            "input_source": input_source,
        },
        "input_type_info": {
            "input_cat_columns": input_cat_columns,
            "input_con_columns": input_con_columns,
        },
    }
    return injections


def _get_output_injections(
    label_file_path: str,
    output_cat_columns: list[str],
    output_con_columns: list[str],
) -> Dict[str, Any]:
    injections = {
        "output_info": {
            "output_source": label_file_path,
        },
        "output_type_info": {
            "target_cat_columns": output_cat_columns,
            "target_con_columns": output_con_columns,
            "uncertainty_weighted_mt_loss": True,
        },
    }

    return injections


def _get_all_dynamic_injections(
    injection_params: ModelInjectionParams,
    genotype_data_path: str,
) -> Dict[str, Any]:
    mip = injection_params

    spe = get_samples_per_epoch(model_injection_params=mip)

    batch_size = get_batch_size(samples_per_epoch=spe.samples_per_epoch)

    valid_size = get_dynamic_valid_size(
        num_samples_per_epoch=spe.samples_per_epoch,
        minimum=batch_size,
    )
    iter_per_epoch = get_num_iter_per_epoch(
        num_samples_per_epoch=spe.samples_per_epoch,
        num_samples_total=spe.num_samples_total,
        batch_size=batch_size,
        valid_size=valid_size,
    )
    bim_path = get_bim_path(genotype_data_path=genotype_data_path)

    n_samples = lines_in_file(file_path=mip.label_file_path) - 1

    subset_folder = Path(mip.modelling_base_output_folder) / "snp_subset_files"

    subset_snp_path = None
    if mip.genotype_feature_selection == "random":
        n_snps, subset_snp_path = build_random_snp_subset_file(
            original_bim_path=Path(bim_path),
            output_folder=subset_folder,
            fold=mip.fold,
            fraction_per_chr=0.1,
        )
    else:
        assert not mip.genotype_feature_selection
        n_snps = lines_in_file(file_path=bim_path)

    injections = {
        "global_config": _get_global_injections(
            fold=mip.fold,
            output_folder=mip.output_folder,
            batch_size=batch_size,
            manual_valid_ids_file=mip.manual_valid_ids_file,
            valid_size=valid_size,
            iter_per_epoch=iter_per_epoch,
            n_snps=n_snps,
            n_samples=n_samples,
            weighted_sampling_columns=mip.weighted_sampling_columns,
            data_format=mip.data_format,
        ),
        "input_genotype_config": _get_genotype_injections(
            input_source=mip.genotype_input_source,
            n_snps=n_snps,
            subset_snp_path=subset_snp_path,
        ),
        "fusion_config": {},
        "output_config": {},
    }

    for output_config in mip.output_configs:
        cat_cols = output_config["output_type_info"]["target_cat_columns"]
        con_cols = output_config["output_type_info"]["target_con_columns"]

        cur_config_name = output_config["output_info"]["output_name"]

        cur_injections = _get_output_injections(
            label_file_path=mip.label_file_path,
            output_cat_columns=cat_cols,
            output_con_columns=con_cols,
        )

        injections["output_config"][cur_config_name] = cur_injections

    if mip.input_cat_columns or mip.input_con_columns:
        injections["input_tabular_config"] = _get_tabular_injections(
            input_source=mip.label_file_path,
            input_cat_columns=list(mip.input_cat_columns),
            input_con_columns=list(mip.input_con_columns),
        )

    return injections


def get_num_iter_per_epoch(
    num_samples_per_epoch: int,
    num_samples_total: int,
    batch_size: int,
    valid_size: int,
) -> int:

    min_iter_per_epoch = 500
    if num_samples_total < 10_000:
        min_iter_per_epoch = 100

    iter_per_epoch = (num_samples_per_epoch - valid_size) // batch_size
    iter_per_epoch = max(min_iter_per_epoch, iter_per_epoch)
    return iter_per_epoch


def build_random_snp_subset_file(
    original_bim_path: Path,
    output_folder: Path,
    fold: int,
    fraction_per_chr: float = 0.1,
) -> tuple[int, Path]:
    df_bim = read_bim(bim_file_path=str(original_bim_path))

    grouped = df_bim.groupby("CHR_CODE")

    sampled_dfs = []
    for _, group in grouped:
        sample_size = int(len(group) * fraction_per_chr)
        sampled_dfs.append(
            group.sample(
                n=sample_size,
                random_state=fold,
                replace=False,
            )
        )

    df_sampled = pd.concat(sampled_dfs).sort_values(["CHR_CODE", "BP_COORD"])

    output_file = output_folder / f"random_subset_fold={fold}.txt"

    if output_file.exists():
        logger.info("%s already exists, using file.", output_file)
        n_snps = lines_in_file(file_path=output_file)
        return n_snps, output_file

    ensure_path_exists(path=output_file.parent, is_folder=True)

    df_sampled = df_sampled[["VAR_ID"]]

    df_sampled.to_csv(
        output_file,
        sep="\t",
        header=False,
        index=False,
    )

    logger.info(
        "Created random SNP subset file with %d SNPs: %s for fold %d",
        len(df_sampled),
        output_file,
        fold,
    )

    return len(df_sampled), output_file



import shutil
import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from typing import Optional

import pandas as pd
from aislib.misc_utils import ensure_path_exists, get_logger
from eir.setup.input_setup_modules.setup_omics import read_bim
from rich.console import Console
from rich.progress import Progress
from rich.table import Table

logger = get_logger(name=__name__)


def run_sync(
    genotype_data_path: Path,
    experiment_folder: Path,
    data_output_folder: Path,
    output_folder: Path,
) -> str:
    with Progress() as progress:
        genotype_base_name = check_genotype_folder(genotype_folder=genotype_data_path)

        df_bim_exp = get_experiment_bim_file(experiment_folder=experiment_folder)
        df_bim_prd = get_predict_bim_file(genotype_folder=genotype_data_path)

        log_output_path = data_output_folder / "snp_overlap_analysis.txt"
        log_overlap(
            df_bim_prd=df_bim_prd,
            df_bim_exp=df_bim_exp,
            output_path=log_output_path,
        )

        sync_task = progress.add_task(
            description="[green]Synchronizing genotype data...",
            total=4,
        )

        cols = ["CHR_CODE", "BP_COORD", "ALT", "REF"]
        df_bim_prd["key"] = df_bim_prd[cols].astype(str).agg("-".join, axis=1)
        df_bim_exp["key"] = df_bim_exp[cols].astype(str).agg("-".join, axis=1)
        to_remove = df_bim_prd[~df_bim_prd["key"].isin(df_bim_exp["key"])]
        df_remove = df_bim_prd.loc[to_remove.index]
        df_remove = df_remove.drop(columns=["key"])

        progress.advance(sync_task)

        filtered_genotype_data_path = remove_extra_snps(
            df_remove=df_remove,
            genotype_data_path=str(genotype_data_path),
            genotype_base_name=genotype_base_name,
            output_folder=output_folder,
        )
        progress.advance(sync_task)

        to_add = df_bim_exp[~df_bim_exp["key"].isin(df_bim_prd["key"])]
        df_add = df_bim_exp.loc[to_add.index]

        added_genotype_data = add_missing_snps(
            df_add=df_add,
            genotype_data_path=filtered_genotype_data_path,
            genotype_base_name=genotype_base_name,
            output_folder=data_output_folder,
        )
        progress.advance(sync_task)

        reordered_genotype_data = update_and_reorder(
            genotype_input_folder=added_genotype_data,
            df_exp_bim=df_bim_exp,
            output_folder=data_output_folder,
        )
        progress.advance(sync_task)

        df_bim_exp = df_bim_exp.drop(columns=["key"])
        df_bim_final = get_predict_bim_file(
            genotype_folder=Path(reordered_genotype_data)
        )
        assert df_bim_final[cols].equals(df_bim_exp[cols])
        progress.advance(sync_task)

    print("\n")

    return str(reordered_genotype_data)


def log_overlap(
    df_bim_prd: pd.DataFrame,
    df_bim_exp: pd.DataFrame,
    output_path: Optional[str | Path] = None,
) -> None:
    logger.info(
        "Examining SNP overlap between current prediction and previous experiment."
    )

    console = Console()

    overlap = df_bim_prd.merge(df_bim_exp, on=["CHR_CODE", "BP_COORD", "ALT", "REF"])
    prd_count = len(df_bim_prd)
    exp_count = len(df_bim_exp)
    overlap_count = len(overlap)
    to_remove_count = prd_count - overlap_count
    to_add_count = exp_count - overlap_count
    overlap_percentage = (overlap_count / prd_count * 100) if prd_count > 0 else 0

    color = (
        "green"
        if overlap_percentage > 75
        else "yellow" if overlap_percentage > 50 else "red"
    )

    table = Table(title="SNP Overlap Analysis", title_style="bold blue")
    table.add_column("Metric", style="dim", width=40)
    table.add_column("Value", style="bold")

    results = {
        "overlapping_snps": overlap_count,
        "overlap_percentage": f"{overlap_percentage:.2f}%",
        "previous_experiment_snps": exp_count,
        "current_prediction_snps": prd_count,
        "snps_to_remove": to_remove_count,
        "snps_to_add": to_add_count,
    }

    table.add_row("Number of overlapping SNPs", f"[{color}]{overlap_count:,}[/]")
    table.add_row(
        "Percentage of current prediction SNPs overlapping",
        f"[{color}]{overlap_percentage:.2f}%[/]",
    )
    table.add_row("Number of SNPs in previous experiment", f"{exp_count:,}")
    table.add_row("Number of SNPs for current prediction", f"{prd_count:,}")
    table.add_row("Number of SNPs to remove", f"{to_remove_count:,}")
    table.add_row("Number of SNPs to add", f"{to_add_count:,}")

    console.print("\n")
    console.print(table)
    console.print("\n")

    if output_path:
        try:
            overlap_quality = (
                "High"
                if overlap_percentage > 75
                else "Medium" if overlap_percentage > 50 else "Low"
            )
            results_formatted = [
                f"{k.replace('_', ' ').title()}: {v}" for k, v in results.items()
            ]
            output_text = [
                "SNP Overlap Analysis Results",
                "=========================\n",
                *results_formatted,
                "\nDetailed Information:",
                f"- The overlap between current prediction and previous experiment"
                f" is {overlap_percentage:.2f}%",
                f"- {to_remove_count:,} SNPs need to be removed from the current "
                f"prediction",
                f"- {to_add_count:,} SNPs need to be added from the previous "
                f"experiment",
                "\nQuality Assessment:",
                f"- Overlap Quality: {overlap_quality}",
            ]

            with open(output_path, "w") as f:
                f.write("\n".join(output_text))

            logger.info(f"Results saved to {output_path}")
        except Exception as e:
            logger.error(f"Failed to save results to {output_path}: {str(e)}")


def check_genotype_folder(genotype_folder: Path) -> str:
    files = [
        i for i in genotype_folder.iterdir() if i.suffix in [".bed", ".bim", ".fam"]
    ]
    assert len(files) == 3

    name = files[0].stem
    assert all(i.stem == name for i in files)

    return name


def remove_extra_snps(
    df_remove: pd.DataFrame,
    genotype_data_path: str,
    genotype_base_name: str,
    output_folder: Path,
) -> str:
    if len(df_remove) == 0:
        return genotype_data_path

    filtered_path = output_folder / "filtered_genotype_data"
    filtered_path.mkdir(exist_ok=True, parents=True)

    snps_to_remove_filepath = str(filtered_path / "snps_to_remove.txt")

    df_remove[["VAR_ID"]].to_csv(
        snps_to_remove_filepath,
        index=False,
        header=False,
        sep="\t",
    )

    remove_command = [
        "plink2",
        "--bfile",
        genotype_data_path + "/" + genotype_base_name,
        "--exclude",
        snps_to_remove_filepath,
        "--keep-allele-order",
        "--make-bed",
        "--out",
        filtered_path / genotype_base_name,
    ]
    run_subprocess(command=remove_command)

    return str(filtered_path)


def add_missing_snps(
    df_add: pd.DataFrame,
    genotype_data_path: str,
    genotype_base_name: str,
    output_folder: Path,
) -> str:
    if len(df_add) == 0:
        return genotype_data_path

    num_snps_to_add = len(df_add)

    fam_file_path = Path(genotype_data_path) / f"{genotype_base_name}.fam"
    num_samples = sum(1 for _ in open(fam_file_path))

    dummy_path = output_folder / "dummy_genotype_data"
    dummy_path.mkdir(exist_ok=True, parents=True)

    dummy_command = [
        "plink2",
        "--dummy",
        str(num_samples),
        str(num_snps_to_add),
        "1",
        "--out",
        str(dummy_path / "dummy_dataset"),
        "--make-bed",
    ]
    run_subprocess(command=dummy_command)

    bim_file_path = dummy_path / "dummy_dataset.bim"
    if bim_file_path.exists():
        bim_file_path.unlink()

    df_add.to_csv(bim_file_path, index=False, header=False, sep="\t")

    fam_file_destination = dummy_path / "dummy_dataset.fam"
    shutil.copy(fam_file_path, fam_file_destination)

    output_folder_missing_imputed = output_folder / "with_missing"
    ensure_path_exists(path=output_folder_missing_imputed, is_folder=True)

    merge_command = [
        "plink",
        "--bfile",
        str(Path(genotype_data_path) / genotype_base_name),
        "--bmerge",
        str(dummy_path / "dummy_dataset"),
        "--keep-allele-order",
        "--make-bed",
        "--out",
        str(output_folder_missing_imputed / (genotype_base_name + "_with_missing")),
    ]
    run_subprocess(command=merge_command)

    return str(output_folder_missing_imputed)


def get_experiment_bim_file(experiment_folder: Path) -> pd.DataFrame:
    bim_file = experiment_folder / "meta" / "snps.bim"
    assert bim_file.exists()

    df_bim_experiment = read_bim_and_cast_dtypes(bim_file_path=str(bim_file))

    return df_bim_experiment


def get_predict_bim_file(genotype_folder: Path) -> pd.DataFrame:
    bim_files = [i for i in genotype_folder.iterdir() if i.suffix == ".bim"]
    assert len(bim_files) == 1

    bim_file = bim_files[0]

    df_bim_predict = read_bim_and_cast_dtypes(bim_file_path=str(bim_file))

    return df_bim_predict


def update_and_reorder(
    genotype_input_folder: str,
    df_exp_bim: pd.DataFrame,
    output_folder: Path,
) -> Path:
    genotype_input_path = Path(genotype_input_folder)
    output_path = Path(output_folder, "reordered_genotype_data")
    output_path.mkdir(parents=True, exist_ok=True)

    bim_files = [i for i in genotype_input_path.iterdir() if i.suffix == ".bim"]
    assert len(bim_files) == 1, "There should be exactly one .bim file in the folder."
    bim_file = bim_files[0]
    df_bim_original = read_bim_and_cast_dtypes(bim_file_path=str(bim_file))
    df_bim = df_bim_original.copy()

    mapping = df_exp_bim.set_index(
        [
            "CHR_CODE",
            "BP_COORD",
            "ALT",
            "REF",
        ]
    )["VAR_ID"].to_dict()

    df_bim["VAR_ID"] = df_bim.apply(
        lambda row: mapping.get(
            (row["CHR_CODE"], row["BP_COORD"], row["ALT"], row["REF"]), row["VAR_ID"]
        ),
        axis=1,
    )
    df_bim.to_csv(bim_file, sep="\t", header=False, index=False)

    merged_output_path = create_and_merge_dummy_fileset(
        genotype_input_folder=genotype_input_folder,
        df_exp_bim=df_exp_bim,
        output_folder=output_path,
    )

    df_bim_original.to_csv(bim_file, sep="\t", header=False, index=False)

    return merged_output_path


def create_and_merge_dummy_fileset(
    genotype_input_folder: str,
    df_exp_bim: pd.DataFrame,
    output_folder: Path,
) -> Path:
    genotype_input_path = Path(genotype_input_folder)
    output_folder.mkdir(parents=True, exist_ok=True)

    base_name = check_genotype_folder(genotype_folder=Path(genotype_input_folder))
    num_snps = len(df_exp_bim)

    dummy_path = output_folder / "dummy_genotype_data"
    dummy_path.mkdir(exist_ok=True, parents=True)
    dummy_command = [
        "plink2",
        "--dummy",
        "1",
        str(num_snps),
        "1",
        "--out",
        str(dummy_path / "dummy_dataset"),
        "--make-bed",
    ]
    run_subprocess(command=dummy_command)

    bim_file_path = dummy_path / "dummy_dataset.bim"
    df_exp_bim[
        [
            "CHR_CODE",
            "VAR_ID",
            "POS_CM",
            "BP_COORD",
            "ALT",
            "REF",
        ]
    ].to_csv(bim_file_path, sep="\t", header=False, index=False)

    merged_output_path = output_folder / f"{base_name}_merged"
    merge_command = [
        "plink",
        "--bfile",
        str(genotype_input_path / base_name),
        "--bmerge",
        str(dummy_path / "dummy_dataset"),
        "--keep-allele-order",
        "--make-bed",
        "--out",
        str(merged_output_path),
    ]
    run_subprocess(command=merge_command)

    remove_command = [
        "plink",
        "--bfile",
        str(merged_output_path),
        "--remove",
        str(dummy_path / "dummy_dataset.fam"),
        "--keep-allele-order",
        "--make-bed",
        "--out",
        str(output_folder / f"{base_name}_reordered"),
    ]
    run_subprocess(command=remove_command)

    shutil.rmtree(output_folder / "dummy_genotype_data")

    keep_name = f"{base_name}_reordered"
    for f in output_folder.iterdir():
        if f.is_file() and not f.stem.startswith(keep_name):
            f.unlink()

    return output_folder


def run_subprocess(command: list[str]) -> None:
    try:
        subprocess.run(
            command,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
    except CalledProcessError as e:
        logger.error(f"Command failed with exit code {e.returncode}")
        logger.error(f"Output: {e.output}")
        logger.error(f"Error: {e.stderr}")
        raise


def read_bim_and_cast_dtypes(bim_file_path: Path | str) -> pd.DataFrame:

    dtypes = {
        "CHR_CODE": str,
        "VAR_ID": str,
        "POS_CM": float,
        "BP_COORD": int,
        "ALT": str,
        "REF": str,
    }

    df_bim = read_bim(bim_file_path=str(bim_file_path))

    df_bim = df_bim.astype(dtypes)

    return df_bim
